#!/usr/bin/env python
# node.py
#
# Copyright (C) 2007-2008 Francois Aucamp, Meraka Institute, CSIR
# See AUTHORS for all authors and contact information.
#
# License: GNU Lesser General Public License, version 3 or later; see COPYING
#          included in this archive for details.
#
# This library is free software, distributed under the terms of
# the GNU Lesser General Public License Version 3, or any later version.
# See the COPYING file included in this archive
#
# The docstrings in this module contain epytext markup; API documentation
# may be created by processing this file with epydoc: http://epydoc.sf.net

from __future__ import absolute_import
from __future__ import print_function
import six
from six.moves import range
from io import open

import hashlib
import json
import random
import time
import traceback

from twisted.internet import defer
import twisted.internet.reactor
import twisted.internet.threads

from . import constants  # @UnresolvedImport
from . import routingtable  # @UnresolvedImport
from . import datastore  # @UnresolvedImport
from . import protocol  # @UnresolvedImport
from . import encoding  # @UnresolvedImport
from .contact import Contact, LayeredContact  # @UnresolvedImport

_Debug = False


def rpcmethod(func):
    """
    Decorator to expose Node methods as remote procedure calls.

    Apply this decorator to methods in the Node class (or a subclass) in
    order to make them remotely callable via the DHT's RPC mechanism.
    """
    func.rpcmethod = True
    return func


def storeSuccess(ok, key):
    try:
        if isinstance(ok, six.binary_type):
            try:
                ok = ok.decode()
            except:
                ok = ok.decode(errors='ignore')
        ok = str(ok)
    except:
        ok = 'Unknown Error'
    if _Debug:
        print('[DHT NODE]    storeSuccess', key, ok)
    return ok


def getErrorMessage(x):
    try:
        errmsg = x.value.subFailure.getErrorMessage()
    except:
        try:
            errmsg = x.getErrorMessage()
        except:
            try:
                errmsg = x.value
            except:
                try:
                    errmsg = str(x)
                except:
                    errmsg = 'Unknown Error'
    if isinstance(errmsg, six.binary_type):
        try:
            errmsg = errmsg.decode()
        except:
            errmsg = errmsg.decode(errors='ignore')
    return errmsg


def storeFailed(x, key):
    errmsg = getErrorMessage(x)
    if _Debug:
        print('[DHT NODE]    storeFailed', key, errmsg, x)
    return errmsg


def findNodeFailed(x):
    errmsg = getErrorMessage(x)
    if _Debug:
        print('[DHT NODE]    findNodeFailed', errmsg, x)
    return errmsg


def lookupFailed(x):
    errmsg = getErrorMessage(x)
    if _Debug:
        print('[DHT NODE]    lookupFailed', errmsg, x)
    return errmsg


class Node(object):
    """
    Local node in the Kademlia network.

    This class represents a single local node in a Kademlia network; in other
    words, this class encapsulates an Entangled-using application's "presence"
    in a Kademlia network.

    In Entangled, all interactions with the Kademlia network by a client
    application is performed via this class (or a subclass).
    """
    def __init__(self, udpPort=4000, dataStore=None, routingTable=None, networkProtocol=None, **kwargs):
        """
        @param dataStore: The data store to use. This must be class inheriting
                          from the C{DataStore} interface (or providing the
                          same API). How the data store manages its data
                          internally is up to the implementation of that data
                          store.
        @type dataStore: entangled.kademlia.datastore.DataStore
        @param routingTable: The routing table to use. Since there exists some
                             ambiguity as to how the routing table should be
                             implemented in Kademlia, a different routing table
                             may be used, as long as the appropriate API is
                             exposed.
        @type routingTable: entangled.kademlia.routingtable.RoutingTable
        @param networkProtocol: The network protocol to use. This can be
                                overridden from the default to (for example)
                                change the format of the physical RPC messages
                                being transmitted.
        @type networkProtocol: entangled.kademlia.protocol.KademliaProtocol
        """
        self.id = kwargs.get('id')
        if not self.id:
            self.id = self._generateID()
        self.port = udpPort
        self.listener = None
        self.refresher = None
        # This will contain a deferred created when joining the network, to enable publishing/retrieving information from
        # the DHT as soon as the node is part of the network (add callbacks to this deferred if scheduling such operations
        # before the node has finished joining the network)
        self._joinDeferred = None
        # Create k-buckets (for storing contacts)
        if routingTable is None:
            self._routingTable = routingtable.TreeRoutingTable(self.id)
        else:
            self._routingTable = routingTable

        # Initialize this node's network access mechanisms
        if networkProtocol is None:
            self._protocol = protocol.KademliaProtocol(self)
        else:
            self._protocol = networkProtocol(self)
        # Initialize the data storage mechanism used by this node
        if dataStore is None:
            self._dataStore = datastore.DictDataStore()  # in memory only
        else:
            self._dataStore = dataStore
            # Try to restore the node's state...

            h = hashlib.sha1()
            h.update(b'nodeState')
            nodeStateKey = h.hexdigest()

            if nodeStateKey in self._dataStore:
                json_state = self._dataStore[nodeStateKey]
                state = json.loads(json_state)
                self.id = state['id']
                for contactTriple in state['closestNodes']:
                    contact = Contact(encoding.to_text(contactTriple[0]), contactTriple[1], contactTriple[2], self._protocol)
                    self._routingTable.addContact(contact)
                if _Debug:
                    print('[DHT NODE]    found "nodeState" key in local db and added %d contacts to routing table' % len(state['closestNodes']))
        self._counter = None

    def __del__(self):
        self._persistState()

    def listenUDP(self):
        self.listener = twisted.internet.reactor.listenUDP(self.port, self._protocol)  # IGNORE:E1101   @UndefinedVariable

    def joinNetwork(self, knownNodeAddresses=None):
        """
        Causes the Node to join the Kademlia network; normally, this should be
        called before any other DHT operations.

        @param knownNodeAddresses: A sequence of tuples containing IP address
                                   information for existing nodes on the
                                   Kademlia network, in the format:
                                   C{(<ip address>, (udp port>)}
        @type knownNodeAddresses: tuple
        """
        # if self._counter:
        #     self._counter('joinNetwork')
        # Prepare the underlying Kademlia protocol
        # Create temporary contact information for the list of addresses of known nodes
        if knownNodeAddresses is not None:
            bootstrapContacts = []
            for address, port in knownNodeAddresses:
                contact = Contact(self._generateID(), address, port, self._protocol)
                bootstrapContacts.append(contact)
        else:
            bootstrapContacts = None
        # Initiate the Kademlia joining sequence - perform a search for this node's own ID
        self._joinDeferred = self._iterativeFind(self.id, bootstrapContacts)
        #        #TODO: Refresh all k-buckets further away than this node's closest neighbour
        #        def getBucketAfterNeighbour(*args):
        #            for i in range(160):
        #                if len(self._buckets[i]) > 0:
        #                    return i+1
        #            return 160
        #        df.addCallback(getBucketAfterNeighbour)
        #        df.addCallback(self._refreshKBuckets)
        #protocol.reactor.callLater(10, self.printContacts)
        self._joinDeferred.addCallback(self._persistState)
        self._joinDeferred.addErrback(self._joinNetworkFailed)
        # Start refreshing k-buckets periodically, if necessary
        self.refresher = twisted.internet.reactor.callLater(constants.checkRefreshInterval, self._refreshNode)  # IGNORE:E1101  @UndefinedVariable
        # twisted.internet.reactor.run()
        return self._joinDeferred

    def printContacts(self):
        if _Debug:
            print('\n\nNODE CONTACTS\n===============')
        for i in range(len(self._routingTable._buckets)):
            for contact in self._routingTable._buckets[i]._contacts:
                if _Debug:
                    print(contact)
        if _Debug:
            print('==================================')
        #twisted.internet.reactor.callLater(10, self.printContacts)

    def iterativeStore(self, key, value, originalPublisherID=None, age=0, expireSeconds=constants.dataExpireSecondsDefaut, **kwargs):
        """
        The Kademlia store operation.

        Call this to store/republish data in the DHT.

        @param key: The hashtable key of the data
        @type key: str
        @param value: The actual data (the value associated with C{key})
        @type value: str
        @param originalPublisherID: The node ID of the node that is the
                                    B{original} publisher of the data
        @type originalPublisherID: str
        @param age: The relative age of the data (time in seconds since it was
                    originally published). Note that the original publish time
                    isn't actually given, to compensate for clock skew between
                    different nodes.
        @type age: int
        """
        # if self._counter:
        #     self._counter('iterativeStore')
        if originalPublisherID is None:
            originalPublisherID = self.id
        collect_results = kwargs.pop('collect_results', False)
        ret = defer.Deferred()

        # Prepare a callback for doing "STORE" RPC calls

        def storeRPCsCollected(store_results, store_nodes):
            if _Debug:
                print('[DHT NODE]    storeRPCsCollected', store_results, store_nodes)
            ret.callback((
                store_nodes,
                store_results,
            ))
            return None

        def storeRPCsFailed(x):
            try:
                errmsg = x.value.subFailure.getErrorMessage()
            except:
                try:
                    errmsg = x.getErrorMessage()
                except:
                    try:
                        errmsg = x.value
                    except:
                        try:
                            errmsg = str(x)
                        except:
                            errmsg = 'Unknown Error'
            if isinstance(errmsg, six.binary_type):
                try:
                    errmsg = errmsg.decode()
                except:
                    errmsg = errmsg.decode(errors='ignore')
            if _Debug:
                print('[DHT NODE]    storeRPCsFailed', errmsg)
            ret.errback(x)
            return errmsg

        def executeStoreRPCs(nodes):
            l = []
            if len(nodes) >= constants.k:
                # If this node itself is closer to the key than the last (furthest) node in the list,
                # we should store the value at ourselves as well
                if self._routingTable.distance(key, self.id) < self._routingTable.distance(key, nodes[-1].id):
                    nodes.pop()
                    try:
                        ok = self.store(key, value, originalPublisherID=originalPublisherID, age=age, expireSeconds=expireSeconds, **kwargs)
                        l.append(defer.succeed(ok))
                    except Exception as exc:
                        if _Debug:
                            traceback.print_exc()
                        l.append(defer.fail(exc))
            else:
                try:
                    ok = self.store(key, value, originalPublisherID=originalPublisherID, age=age, expireSeconds=expireSeconds, **kwargs)
                    l.append(defer.succeed(ok))
                except Exception as exc:
                    if _Debug:
                        traceback.print_exc()
                    l.append(defer.fail(exc))

            for contact in nodes:
                d = contact.store(key, value, originalPublisherID, age, expireSeconds, **kwargs)
                d.addCallback(storeSuccess, key)
                d.addErrback(storeFailed, key)
                l.append(d)

            if not collect_results:
                return nodes

            dl = defer.DeferredList(l, fireOnOneErrback=True, consumeErrors=True)
            dl.addCallback(storeRPCsCollected, nodes)
            dl.addErrback(storeRPCsFailed)
            return dl

        # Find k nodes closest to the key...
        df = self.iterativeFindNode(key)
        # ...and send them STORE RPCs as soon as they've been found
        df.addCallback(executeStoreRPCs)
        df.addErrback(findNodeFailed)

        if not collect_results:
            return df

        return ret

    def iterativeFindNode(self, key):
        """
        The basic Kademlia node lookup operation.

        Call this to find a remote node in the P2P overlay network.

        @param key: the 160-bit key (i.e. the node or value ID) to search for
        @type key: str

        @return: This immediately returns a deferred object, which will return
                 a list of k "closest" contacts (C{kademlia.contact.Contact}
                 objects) to the specified key as soon as the operation is
                 finished.
        @rtype: twisted.internet.defer.Deferred
        """
        return self._iterativeFind(key)

    def iterativeFindValue(self, key, rpc='findValue', refresh_revision=False):
        """
        The Kademlia search operation (deterministic)

        Call this to retrieve data from the DHT.

        @param key: the 160-bit key (i.e. the value ID) to search for
        @type key: str

        @return: This immediately returns a deferred object, which will return
                 either one of two things:
                     - If the value was found, it will return a Python
                     dictionary containing the searched-for key (the C{key}
                     parameter passed to this method), and its associated
                     value, in the format:
                     C{<str>key: <str>data_value}
                     - If the value was not found, it will return a list of k
                     "closest" contacts (C{kademlia.contact.Contact} objects)
                     to the specified key
        @rtype: twisted.internet.defer.Deferred
        """
        # Prepare a callback for this operation
        outerDf = defer.Deferred()

        def refreshRevisionSuccess(ok):
            if _Debug:
                print('[DHT NODE]    iterativeFindValue.refreshRevisionSuccess', ok)

        def refreshRevisionFailed(x):
            errmsg = getErrorMessage(x)
            if _Debug:
                print('[DHT NODE]    iterativeFindValue.refreshRevisionFailed', errmsg)
            return errmsg

        def checkResult(result):
            if _Debug:
                print('[DHT NODE]    iterativeFindValue.checkResult key=%s' % key, result)
            if isinstance(result, dict):
                if key in result:
                    latest_revision = 0
                    for v in result['values']:
                        if v[1] > latest_revision:
                            latest_revision = v[1]
                    # We have found the value; now see who was the closest contact without it...
                    if 'closestNodeNoValue' in result:
                        # ...and store the key/value pair
                        contact = result['closestNodeNoValue']
                        expireSeconds = constants.dataExpireSecondsDefaut
                        if 'expireSeconds' in result:
                            expireSeconds = result['expireSeconds']
                        if _Debug:
                            print('[DHT NODE]    republish %s to closest node with %d expire seconds' % (key, expireSeconds))
                        contact.store(key, result[key], None, 0, expireSeconds, revision=latest_revision).addErrback(storeFailed)
                    if refresh_revision:
                        # need to refresh nodes who has old version of that value
                        for v in result['values']:
                            if v[1] < latest_revision:
                                _contact = Contact(encoding.to_text(v[2]), v[3][0], v[3][1], self._protocol)
                                if _Debug:
                                    print('[DHT NODE]    will refresh revision %d on %r' % (latest_revision, _contact))
                                d = _contact.store(key, result[key], None, 0, expireSeconds, revision=latest_revision)
                                d.addCallback(refreshRevisionSuccess)
                                d.addErrback(refreshRevisionFailed)
                    outerDf.callback(result)
                else:
                    # we was looking for value but did not found it
                    # Now, see if we have the value (it might seem wasteful to search on the network
                    # first, but it ensures that all values are properly propagated through the
                    # network
                    if key in self._dataStore:
                        # Ok, we have the value locally, so use that
                        item = self._dataStore.getItem(key)
                        expireSeconds = item.get('expireSeconds', constants.dataExpireSecondsDefaut)
                        # Send this value to the closest node without it
                        if len(result['activeContacts']) > 0:
                            contact = result['activeContacts'][0]
                            if _Debug:
                                print('[DHT NODE]    refresh %s : %r with %d to %r' % (key, item['value'], expireSeconds, contact))
                            contact.store(key, item['value'], None, 0, expireSeconds).addErrback(storeFailed)
                        outerDf.callback({
                            'key': item['value'],
                            'values': [
                                (
                                    item['value'],
                                    item['revision'],
                                    self.id,
                                    (b'127.0.0.1', self.port),
                                ),
                            ],
                            'activeContacts': result['activeContacts'],
                        })
                    else:
                        if _Debug:
                            print('[DHT NODE]    key %s does not exist in DHT' % key)
                        outerDf.callback(result)
            else:
                # The value wasn't found, but a list of contacts was returned
                # Now, see if we have the value (it might seem wasteful to search on the network
                # first, but it ensures that all values are properly propagated through the
                # network
                if key in self._dataStore:
                    # Ok, we have the value locally, so use that
                    item = self._dataStore.getItem(key)
                    expireSeconds = item.get('expireSeconds', constants.dataExpireSecondsDefaut)
                    # Send this value to the closest node without it
                    if len(result) > 0:
                        contact = result[0]
                        if _Debug:
                            print('[DHT NODE]    refresh %s on closest node : %r with %d to %r' % (key, item['value'], expireSeconds, contact))
                        contact.store(key, item['value'], None, 0, expireSeconds).addErrback(storeFailed)
                    ret = {
                        'key': item['value'],
                        'values': [
                            (
                                item['value'],
                                item['revision'],
                                self.id,
                                (b'127.0.0.1', self.port),
                            ),
                        ],
                        'activeContacts': result['activeContacts'],
                    }
                    if _Debug:
                        print('[DHT NODE]    key %s found in my local data store : %r' % (key, ret))
                    outerDf.callback(ret)
                else:
                    if _Debug:
                        print('[DHT NODE]    key %s does not exist in DHT' % key)
                    outerDf.callback(result)

        # Execute the search
        df = self._iterativeFind(key, rpc=rpc)
        df.addCallback(checkResult)
        df.addErrback(lookupFailed)
        return outerDf

    def addContact(self, contact):
        """
        Add/update the given contact; simple wrapper for the same method in
        this object's RoutingTable object.

        @param contact: The contact to add to this node's k-buckets
        @type contact: kademlia.contact.Contact
        """
        # if self._counter:
        #     self._counter('addContact')
        self._routingTable.addContact(contact)

    def removeContact(self, contactID):
        """
        Remove the contact with the specified node ID from this node's table of
        known nodes. This is a simple wrapper for the same method in this
        object's RoutingTable object.

        @param contactID: The node ID of the contact to remove
        @type contactID: str
        """
        # if self._counter:
        #     self._counter('removeContact')
        self._routingTable.removeContact(contactID)

    def findContact(self, contactID):
        """
        Find a entangled.kademlia.contact.Contact object for the specified
        cotact ID.

        @param contactID: The contact ID of the required Contact object
        @type contactID: str

        @return: Contact object of remote node with the specified node ID
        @rtype: twisted.internet.defer.Deferred
        """
        # if self._counter:
        #     self._counter('findContact')
        try:
            contact = self._routingTable.getContact(contactID)
            df = defer.Deferred()
            df.callback(contact)
        except ValueError:

            def parseResults(nodes):
                if contactID in nodes:
                    contact = nodes[nodes.index(contactID)]
                    return contact
                else:
                    return None

            df = self.iterativeFindNode(contactID)
            df.addCallback(parseResults)
        return df

    @rpcmethod
    def ping(self):
        """
        Used to verify contact between two Kademlia nodes.

        @rtype: str
        """
        # if self._counter:
        #     self._counter('rpc_node_ping')
        return 'pong'

    @rpcmethod
    def store(self, key, value, originalPublisherID=None, age=0, expireSeconds=constants.dataExpireSecondsDefaut, **kwargs):
        """
        Store the received data in this node's local hash table.

        @param key: The hashtable key of the data
        @type key: str
        @param value: The actual data (the value associated with C{key})
        @type value: str
        @param originalPublisherID: The node ID of the node that is the
                                    B{original} publisher of the data
        @type originalPublisherID: str
        @param age: The relative age of the data (time in seconds since it was
                    originally published). Note that the original publish time
                    isn't actually given, to compensate for clock skew between
                    different nodes.
        @type age: int

        @rtype: str

        @todo: Since the data (value) may be large, passing it around as a buffer
               (which is the case currently) might not be a good idea... will have
               to fix this (perhaps use a stream from the Protocol class?)
        """
        # if self._counter:
        #     self._counter('rpc_node_store')
        if _Debug:
            print('[DHT NODE]  SINGLE  rpcmethod.store %r' % key)
        # Get the sender's ID (if any)
        if '_rpcNodeID' in kwargs:
            rpcSenderID = kwargs['_rpcNodeID']
        else:
            rpcSenderID = None

        if originalPublisherID is None:
            if rpcSenderID is not None:
                originalPublisherID = rpcSenderID
            else:
                raise TypeError('No publisher specifed, and RPC caller ID not available. Data requires an original publisher.')

        now = int(time.time())
        originallyPublished = now - age
        self._dataStore.setItem(key, value, now, originallyPublished, originalPublisherID, expireSeconds=expireSeconds, **kwargs)
        return 'OK'

    @rpcmethod
    def findNode(self, key, **kwargs):
        """
        Finds a number of known nodes closest to the node/value with the
        specified key.

        @param key: the 160-bit key (i.e. the node or value ID) to search for
        @type key: str

        @return: A list of contact triples closest to the specified key.
                 This method will return C{k} (or C{count}, if specified)
                 contacts if at all possible; it will only return fewer if the
                 node is returning all of the contacts that it knows of.
        @rtype: list
        """
        # if self._counter:
        #     self._counter('rpc_node_findNode')
        if _Debug:
            print('[DHT NODE]  SINGLE rpcmethod.findNode %r' % key)
        # Get the sender's ID (if any)
        if '_rpcNodeID' in kwargs:
            rpcSenderID = kwargs['_rpcNodeID']
        else:
            rpcSenderID = None
        contacts = self._routingTable.findCloseNodes(key, constants.k, rpcSenderID)
        contactTriples = []
        for contact in contacts:
            contactTriples.append((contact.id, encoding.to_text(contact.address), contact.port))
        return contactTriples

    @rpcmethod
    def findValue(self, key, **kwargs):
        """
        Return the value associated with the specified key if present in this
        node's data, otherwise execute FIND_NODE for the key.

        @param key: The hashtable key of the data to return
        @type key: str

        @return: A dictionary containing the requested key/value pair,
                 or a list of contact triples closest to the requested key.
        @rtype: dict or list
        """
        # if self._counter:
        #     self._counter('rpc_node_findValue')
        if _Debug:
            print('[DHT NODE] SINGLE rpcmethod.findValue %r' % key)
        if key in self._dataStore:
            exp = None
            expireSecondsCall = getattr(self._dataStore, 'expireSeconds')
            if expireSecondsCall:
                exp = expireSecondsCall(key)
            originalPublishTimeCall = getattr(self._dataStore, 'originalPublishTime')
            published = None
            if originalPublishTimeCall:
                published = originalPublishTimeCall(key)
            if _Debug:
                print('[DHT NODE]  SINGLE    found key in local dataStore %r' % self._dataStore[key])
            return {
                key: self._dataStore[key],
                'expireSeconds': exp,
                'originallyPublished': published,
            }
        if _Debug:
            print('[DHT NODE] SINGLE     NOT found key in local dataStore')
        return self.findNode(key, **kwargs)

    def _generateID(self):
        """
        Generates a 160-bit pseudo-random identifier.

        @return: A globally unique 160-bit pseudo-random identifier
        @rtype: str
        """
        hsh = hashlib.sha1()
        hsh.update(str(random.getrandbits(255)).encode())
        return hsh.hexdigest()

    def _iterativeFind(self, key, startupShortlist=None, rpc='findNode', deep=False):
        """
        The basic Kademlia iterative lookup operation (for nodes/values)

        This builds a list of k "closest" contacts through iterative use of
        the "FIND_NODE" RPC, or if C{findValue} is set to C{True}, using the
        "FIND_VALUE" RPC, in which case the value (if found) may be returned
        instead of a list of contacts

        @param key: the 160-bit key (i.e. the node or value ID) to search for
        @type key: str
        @param startupShortlist: A list of contacts to use as the starting
                                 shortlist for this search; this is normally
                                 only used when the node joins the network
        @type startupShortlist: list
        @param rpc: The name of the RPC to issue to remote nodes during the
                    Kademlia lookup operation (e.g. this sets whether this
                    algorithm should search for a data value (if
                    rpc='findValue') or not. It can thus be used to perform
                    other operations that piggy-back on the basic Kademlia
                    lookup operation (Entangled's "delete" RPC, for instance).
        @type rpc: str

        @return: If C{findValue} is C{True}, the algorithm will stop as soon
                 as a data value for C{key} is found, and return a dictionary
                 containing the key and the found value. Otherwise, it will
                 return a list of the k closest nodes to the specified key
        @rtype: twisted.internet.defer.Deferred
        """
        if _Debug:
            print('[DHT NODE]  SINGLE _iterativeFind rpc=%r   key=%r  startupShortlist=%r' % (
                rpc,
                key,
                startupShortlist,
            ))
        # if self._counter:
        #     self._counter('_iterativeFind')
        if rpc != 'findNode':
            findValue = True
        else:
            findValue = False
        shortlist = []
        if startupShortlist is None:
            shortlist = self._routingTable.findCloseNodes(key, constants.alpha)
            if key != self.id:
                # Update the "last accessed" timestamp for the appropriate k-bucket
                self._routingTable.touchKBucket(key)
            if len(shortlist) == 0:
                if _Debug:
                    print('[DHT NODE]  SINGLE  This node does not know of any other nodes !!!!!')
                # This node doesn't know of any other nodes
                fakeDf = defer.Deferred()
                fakeDf.callback([])
                return fakeDf
        else:
            # This is used during the bootstrap process; node ID's are most probably fake
            shortlist = startupShortlist
        if _Debug:
            print('[DHT NODE]  SINGLE  shortlist=%r' % shortlist)
        # List of active queries; len() indicates number of active probes
        # - using lists for these variables, because Python doesn't allow binding a new value to a name in an enclosing (non-global) scope
        activeProbes = []
        # List of contact IDs that have already been queried
        alreadyContacted = []
        # Probes that were active during the previous iteration
        # A list of found and known-to-be-active remote nodes
        activeContacts = []
        # This should only contain one entry; the next scheduled iteration call
        pendingIterationCalls = []
        prevClosestNode = [None]
        findValueResult = {
            'values': [],
        }
        slowNodeCount = [0]

        def extendShortlist(responseTuple):
            """ @type responseMsg: kademlia.msgtypes.ResponseMessage """
            # The "raw response" tuple contains the response message, and the originating address info
            responseMsg = responseTuple[0]
            originAddress = responseTuple[1]  # tuple: (ip adress, udp port)
            # Make sure the responding node is valid, and abort the operation if it isn't
            if _Debug:
                print('[DHT NODE]   SINGLE   extendShortlist', (responseMsg.nodeID, type(responseMsg.nodeID)))
            if responseMsg.nodeID in activeContacts or responseMsg.nodeID == self.id:
                if _Debug:
                    if responseMsg.nodeID == self.id:
                        print('[DHT NODE]   SINGLE       response from my own node')
                    else:
                        print('[DHT NODE]   SINGLE       response from active contact')
                return responseMsg.nodeID

            # Mark this node as active
            if responseMsg.nodeID in shortlist:
                # Get the contact information from the shortlist...
                aContact = shortlist[shortlist.index(responseMsg.nodeID)]
            else:
                # If it's not in the shortlist; we probably used a fake ID to reach it
                # - reconstruct the contact, using the real node ID this time
                aContact = Contact(encoding.to_text(responseMsg.nodeID), originAddress[0], originAddress[1], self._protocol)
            activeContacts.append(aContact)
            # This makes sure "bootstrap"-nodes with "fake" IDs don't get queried twice
            if responseMsg.nodeID not in alreadyContacted:
                alreadyContacted.append(responseMsg.nodeID)
            # Now grow extend the (unverified) shortlist with the returned contacts
            result = responseMsg.response
            # TODO: some validation on the result (for guarding against attacks)
            # If we are looking for a value, first see if this result is the value
            # we are looking for before treating it as a list of contact triples
            if findValue and isinstance(result, dict):
                # We have found the value
                findValueResult[key] = result[key]
                findValueResult['values'].append((
                    result[key],
                    result.get('revision', 0),
                    responseMsg.nodeID,
                    originAddress,
                ))
                if 'expireSeconds' in result:
                    findValueResult['expireSeconds'] = result['expireSeconds']
            else:
                if findValue:
                    # We are looking for a value, and the remote node didn't have it
                    # - mark it as the closest "empty" node, if it is
                    if 'closestNodeNoValue' in findValueResult:
                        if self._routingTable.distance(key, responseMsg.nodeID) < self._routingTable.distance(key, activeContacts[0].id):
                            findValueResult['closestNodeNoValue'] = aContact
                    else:
                        findValueResult['closestNodeNoValue'] = aContact
                for contactTriple in result:
                    try:
                        testContact = Contact(encoding.to_text(contactTriple[0]), contactTriple[1], contactTriple[2], self._protocol)
                    except:
                        continue
                    if testContact not in shortlist:
                        shortlist.append(testContact)
            return responseMsg.nodeID

        def removeFromShortlist(failure):
            """ @type failure: twisted.python.failure.Failure """
            failure.trap(protocol.TimeoutError)
            deadContactID = failure.getErrorMessage()
            if deadContactID in shortlist:
                if _Debug:
                    print('[DHT NODE]  SINGLE removing %r' % deadContactID)
                shortlist.remove(deadContactID)
            return deadContactID

        def cancelActiveProbe(contactID):
            activeProbes.pop()
            if len(activeProbes) <= int(constants.alpha/2.0) and len(pendingIterationCalls):
                # Force the iteration
                pendingIterationCalls[0].cancel()
                del pendingIterationCalls[0]
                if _Debug:
                    print('[DHT NODE] SINGLE forcing iteration =================')
                searchIteration()

        # Send parallel, asynchronous FIND_NODE RPCs to the shortlist of contacts
        def searchIteration():
            slowNodeCount[0] = len(activeProbes)
            # Sort the discovered active nodes from closest to furthest
            activeContacts.sort(key=lambda cont: self._routingTable.distance(cont.id, key))
            if _Debug:
                print('[DHT NODE] SINGLE ==> searchiteration %r' % activeContacts)
            # This makes sure a returning probe doesn't force calling this function by mistake
            while len(pendingIterationCalls):
                del pendingIterationCalls[0]
            # See if should continue the search
            if key in findValueResult and not deep:
                if _Debug:
                    print('[DHT NODE] SINGLE ++++++++++++++ DONE (findValue found) +++++++++++++++\n\n')
                findValueResult['activeContacts'] = activeContacts
                outerDf.callback(findValueResult)
                return
            if len(activeContacts) and findValue == False:
                if (len(activeContacts) >= constants.k) or (activeContacts[0] == prevClosestNode[0] and len(activeProbes) == slowNodeCount[0]):
                    # TODO: Re-send the FIND_NODEs to all of the k closest nodes not already queried
                    # Ok, we're done; either we have accumulated k active contacts or no improvement in closestNode has been noted
                    if len(activeContacts) >= constants.k:
                        if _Debug:
                            print('[DHT NODE] SINGLE ++++++++++++++ DONE (test for k active contacts) +++++++++++++++\n\n')
                    else:
                        if _Debug:
                            print('[DHT NODE] SINGLE ++++++++++++++ DONE (test for closest node) +++++++++++++++\n\n')
                    if findValue:
                        findValueResult['activeContacts'] = activeContacts
                        outerDf.callback(findValueResult)
                    else:
                        outerDf.callback(activeContacts)
                    return
            # The search continues...
            if len(activeContacts):
                prevClosestNode[0] = activeContacts[0]
            contactedNow = 0
            activeContacts.sort(key=lambda cont: self._routingTable.distance(cont.id, key))
            # Store the current shortList length before contacting other nodes
            prevShortlistLength = len(shortlist)
            for contact in shortlist:
                if contact.id not in alreadyContacted:
                    activeProbes.append(contact.id)
                    rpcMethod = getattr(contact, rpc)
                    df = rpcMethod(key, rawResponse=True)
                    df.addCallback(extendShortlist)
                    df.addErrback(removeFromShortlist)
                    df.addCallback(cancelActiveProbe)
                    alreadyContacted.append(contact.id)
                    contactedNow += 1
                if contactedNow == constants.alpha:
                    break
            if len(activeProbes) > slowNodeCount[0] \
                    or (len(shortlist) < constants.k and len(activeContacts) < len(shortlist) and len(activeProbes) > 0):
                if _Debug:
                    print('[DHT NODE] SINGLE ----------- scheduling next call -------------')
                # Schedule the next iteration if there are any active calls (Kademlia uses loose parallelism)
                call = twisted.internet.reactor.callLater(constants.iterativeLookupDelay, searchIteration)  # IGNORE:E1101  @UndefinedVariable
                pendingIterationCalls.append(call)
            # Check for a quick contact response that made an update to the shortList
            elif prevShortlistLength < len(shortlist):
                # Ensure that the closest contacts are taken from the updated shortList
                searchIteration()
            else:
                if _Debug:
                    print('[DHT NODE] SINGLE ++++++++++++++ DONE (logically) +++++++++++++\n\n')
                # If no probes were sent, there will not be any improvement, so we're done
                if findValue:
                    findValueResult['activeContacts'] = activeContacts
                    outerDf.callback(findValueResult)
                else:
                    outerDf.callback(activeContacts)

        outerDf = defer.Deferred()
        # Start the iterations
        searchIteration()
        return outerDf

    def _persistState(self, *args):
        state = {
            'id': self.id,
            'closestNodes': self.findNode(self.id),
            'key': 'nodeState',
            'type': 'skip_validation',
        }
        if _Debug:
            print('[DHT NODE]  SINGLE _persistState id=%r state=%r' % (
                self.id,
                state,
            ))
        json_value = json.dumps(state)
        now = int(time.time())

        h = hashlib.sha1()
        h.update(b'nodeState')
        nodeStateKey = h.hexdigest()

        self._dataStore.setItem(nodeStateKey, json_value, now, now, self.id)
        return args

    def _joinNetworkFailed(self, err, **kwargs):
        if _Debug:
            print('[DHT NODE] SINGLE  failed joining DHT network', err)

    def _refreshNode(self):
        """
        Periodically called to perform k-bucket refreshes and data
        replication/republishing as necessary.
        """
        # if self._counter:
        #     self._counter('_refreshNode')
        df = self._refreshRoutingTable()
        df.addCallback(self._republishData)
        df.addCallback(self._scheduleNextNodeRefresh)

    def _refreshRoutingTable(self):
        nodeIDs = self._routingTable.getRefreshList(0, False)
        if _Debug:
            print('[DHT NODE] SINGLE _refreshRoutingTable', nodeIDs)

        outerDf = defer.Deferred()

        def searchFailed(err):
            if _Debug:
                print('[DHT NODE] SINGLE searchFailed', err)

        def searchForNextNodeID(dfResult=None):
            if len(nodeIDs) > 0:
                searchID = nodeIDs.pop()
                df = self.iterativeFindNode(searchID)
                df.addCallback(searchForNextNodeID)
                df.addErrback(searchFailed)
            else:
                # If this is reached, we have finished refreshing the routing table
                outerDf.callback(None)

        # Start the refreshing cycle
        searchForNextNodeID()
        return outerDf

    def _republishData(self, *args):
        df = twisted.internet.threads.deferToThread(self._threadedRepublishData)
        return df

    def _scheduleNextNodeRefresh(self, *args):
        self.refresher = twisted.internet.reactor.callLater(constants.checkRefreshInterval, self._refreshNode)  # @UndefinedVariable

    def _threadedRepublishData(self, *args):
        """
        Republishes and expires any stored data (i.e. stored C{(key, value
        pairs)} that need to be republished/expired.

        This method should run in a deferred thread
        """
        if _Debug:
            print('[DHT NODE]  SINGLE republishData called, node: %r' % self.id)
        expiredKeys = []
        for key in self._dataStore.keys():
            if _Debug:
                print('[DHT NODE]  SINGLE    %r' % key)
            # Filter internal variables stored in the datastore
            if key == 'nodeState':
                continue

            now = int(time.time())
            itemData = self._dataStore.getItem(key)
            originallyPublished = itemData['originallyPublished']
            originalPublisherID = itemData['originalPublisherID']
            lastPublished = itemData['lastPublished']
            expireSeconds = itemData['expireSeconds']
            age = now - originallyPublished
            if originalPublisherID == self.id:
                # This node is the original publisher; it has to republish
                # the data before it expires (24 hours in basic Kademlia)
                if age >= constants.dataExpireTimeout:
                    twisted.internet.reactor.callFromThread(  # @UndefinedVariable
                        self.iterativeStore,
                        key=key,
                        value=itemData['value'],
                        expireSeconds=expireSeconds,
                    )
            else:
                # This node needs to replicate the data at set intervals,
                # until it expires, without changing the metadata associated with it
                # First, check if the data has expired
                if age >= constants.dataExpireTimeout:
                    # This key/value pair has expired (and it has not been republished by the original publishing node
                    # - remove it
                    expiredKeys.append(key)
                elif now - lastPublished >= constants.replicateInterval:
                    # ...data has not yet expired, and we need to replicate it
                    twisted.internet.reactor.callFromThread(  # @UndefinedVariable
                        self.iterativeStore,
                        key=key,
                        value=self._dataStore[key],
                        originalPublisherID=originalPublisherID,
                        age=age,
                        expireSeconds=expireSeconds,
                    )
        for key in expiredKeys:
            del self._dataStore[key]


class MultiLayerNode(Node):
    def __init__(self, udpPort=4000, dataStores=None, routingTables=None, networkProtocol=None, **kwargs):
        self._counter = None
        self.port = udpPort
        self.listener = None
        h = hashlib.sha1()
        h.update(b'nodeState')
        self.nodeStateKey = h.hexdigest()

        self._routingTables = {}
        self._dataStores = {}
        self.layers = {}
        self.refreshers = {}
        self.active_layers = set()
        self.attached_layers = set()

        self.rpc_calls = {}
        self.rpc_responses = {}
        self.packets_in = {}
        self.packets_out = {}
        self.bytes_in = 0
        self.bytes_out = 0

        # This will contain a deferred created when joining the network, to enable publishing/retrieving information from
        # the DHT as soon as the node is part of the network (add callbacks to this deferred if scheduling such operations
        # before the node has finished joining the network)
        self._joinDeferreds = {}

        # Initialize this node's network access mechanisms
        if networkProtocol is None:
            self._protocol = protocol.KademliaMultiLayerProtocol(self)
        else:
            self._protocol = networkProtocol(self)

        if dataStores:
            for layer_id, dataStore in dataStores.items():
                nodeID = None
                if layer_id == 0:
                    nodeID = kwargs.get('id')
                routingTable = None
                if routingTables:
                    routingTable = routingTables.get(layer_id, None)
                self.createLayer(layer_id, dataStore=dataStore, nodeID=nodeID, routingTable=routingTable)

        if not self._dataStores or 0 not in self._dataStores:
            self.createLayer(0, dataStore=datastore.DictDataStore())

        self.active_layers.add(0)
        self.attachLayer(0)

    def createLayer(self, layer_id, dataStore, nodeID=None, routingTable=None):
        if layer_id in self.layers or layer_id in self._dataStores or layer_id in self._routingTables:
            if _Debug:
                print('[DHT NODE]    createLayer : layer %d already exist' % layer_id)
            return False
        self._dataStores[layer_id] = dataStore
        self.layers[layer_id] = nodeID
        loaded = False
        if self.nodeStateKey in self._dataStores[layer_id]:
            json_state = self._dataStores[layer_id][self.nodeStateKey]
            state = json.loads(json_state)
            self.layers[layer_id] = state['id']
            if layer_id not in self._routingTables:
                self._routingTables[layer_id] = routingTable or routingtable.TreeRoutingTable(self.layers[layer_id], layerID=layer_id)
            for contactTriple in state['closestNodes']:
                contact = LayeredContact(encoding.to_text(contactTriple[0]), contactTriple[1], contactTriple[2], self._protocol, layerID=layer_id)
                self._routingTables[layer_id].addContact(contact)
            if _Debug:
                print('[DHT NODE]    createLayer : layer %d : found "nodeState" key in local db and added %d contacts to routing table' % (
                    layer_id,
                    len(state['closestNodes']),
                ))
            loaded = True
        if not self.layers[layer_id]:
            self.layers[layer_id] = self._generateID()
        if layer_id not in self._routingTables:
            self._routingTables[layer_id] = routingTable or routingtable.TreeRoutingTable(self.layers[layer_id], layerID=layer_id)
#         if layer_id != 0 and not loaded and warmUp:
#             loaded = self.warmUpLayer(layer_id)
        if _Debug:
            print('[DHT NODE]    createLayer : layer %d created,  loaded=%r' % (
                layer_id,
                loaded,
            ))
        return True

    def destroyLayer(self, layer_id):
        if layer_id == 0:
            return False
        if layer_id not in self.layers and layer_id not in self._dataStores and layer_id not in self._routingTables:
            if _Debug:
                print('[DHT NODE]    destroyLayer : layer %d not exist' % layer_id)
            return False
        self.detachLayer(layer_id)
        self.active_layers.discard(layer_id)
        self.layers.pop(layer_id, None)
        self._routingTables.pop(layer_id, None)
        self._dataStores.pop(layer_id, None)
        if _Debug:
            print('[DHT NODE]    destroyLayer : layer %d destroyed' % layer_id)
        return True

    def connectingTask(self, layerID=0):
        if layerID not in self._joinDeferreds:
            return None
        return self._joinDeferreds[layerID]

    def joinNetwork(self, knownNodeAddresses=None, layerID=0, attach=False, parallel_calls=None):
        if knownNodeAddresses is not None:
            bootstrapContacts = []
            for address, port in knownNodeAddresses:
                contact = LayeredContact(self._generateID(), address, port, self._protocol, layerID=layerID)
                bootstrapContacts.append(contact)
        else:
            bootstrapContacts = None
        if _Debug:
            print('[DHT NODE]    joinNetwork bootstrapContacts=%r  layerID=%r  attach=%r' % (bootstrapContacts, layerID, attach))
        self.active_layers.add(layerID)
        if attach:
            self.attachLayer(layerID)
        d = self._iterativeFind(
            key=self.layers[layerID],
            startupShortlist=bootstrapContacts,
            layerID=layerID,
            parallel_calls=parallel_calls,
        )
        d.addCallback(self._persistState, layerID=layerID)
        d.addErrback(self._joinNetworkFailed, layerID=layerID)
        self._joinDeferreds[layerID] = d
        if self.refreshers.get(layerID, None) and not self.refreshers[layerID].called:
            self.refreshers[layerID].cancel()
        self.refreshers[layerID] = twisted.internet.reactor.callLater(  # IGNORE:E1101  @UndefinedVariable
            constants.checkRefreshInterval + float(random.randint(0, 15)),
            self._refreshNode,
            layerID=layerID,
        )
        return self._joinDeferreds[layerID]

    def leaveNetwork(self, layerID):
        if not layerID in self.active_layers:
            return False
        if _Debug:
            print('[DHT NODE]    leaveNetwork layerID=%d' % layerID)
        if layerID != 0:
            self.active_layers.discard(layerID)
            self.detachLayer(layerID)
        if self.refreshers.get(layerID, None) and not self.refreshers[layerID].called:
            self.refreshers[layerID].cancel()
        self.refreshers.pop(layerID, None)
        return True

    def attachLayer(self, layerID):
        if _Debug:
            print('[DHT NODE]    attachLayer %d' % layerID)
        self.attached_layers.add(layerID)

    def detachLayer(self, layerID):
        if _Debug:
            print('[DHT NODE]    detachLayer %d' % layerID)
        self.attached_layers.discard(layerID)


#     def warmUpLayer(self, layerID):
#         if layerID == 0:
#             return True
#         if layerID not in self._routingTables:
#             return False
#         json_state = None
#         if self.nodeStateKey in self._dataStores[layerID]:
#             json_state = self._dataStores[layerID][self.nodeStateKey]
#         if not json_state and self.nodeStateKey in self._dataStores[0]:
#             json_state = self._dataStores[0][self.nodeStateKey]
#         if not json_state:
#             return False
#         state = json.loads(json_state)
#         if _Debug:
#             print('[DHT NODE]    warmUpLayer %d   found saved data: %r' % (layerID, state, ))
#         if not state['closestNodes']:
#             return False
#         self.layers[layerID] = state['id']
#         for contactTriple in state['closestNodes']:
#             contact = LayeredContact(encoding.to_text(contactTriple[0]), contactTriple[1], contactTriple[2], self._protocol, layerID=layerID)
#             self._routingTables[layerID].addContact(contact)
#         self.warm_layers.add(layerID)
#         if _Debug:
#             print('[DHT NODE]    warmUpLayer %d : found "nodeState" key in local db of layer 0 and added %d contacts to routing table' % (
#                 layerID, len(state['closestNodes']), ))
#         return True

    def iterativeStore(self, key, value, originalPublisherID=None, age=0, expireSeconds=constants.dataExpireSecondsDefaut, layerID=0, **kwargs):
        if originalPublisherID is None:
            originalPublisherID = self.layers[layerID]
        collect_results = kwargs.pop('collect_results', False)
        parallel_calls = kwargs.pop('parallel_calls', None)
        ret = defer.Deferred()
        if 'layerID' not in kwargs:
            kwargs['layerID'] = layerID

        if _Debug:
            print('[DHT NODE]   iterativeStore layerID=%d  key=%r' % (
                layerID,
                key,
            ))
        # Prepare a callback for doing "STORE" RPC calls

        def storeRPCsCollected(store_results, store_nodes):
            if _Debug:
                print('[DHT NODE]    iterativeStore.storeRPCsCollected', store_results, store_nodes)
            ret.callback((
                store_nodes,
                store_results,
            ))
            return None

        def storeRPCsFailed(x):
            errmsg = storeFailed(x)
            if _Debug:
                print('[DHT NODE]    iterativeStore.storeRPCsFailed', errmsg)
            ret.errback(x)
            return errmsg

        def executeStoreRPCs(nodes):
            l = []
            if len(nodes) >= constants.k:
                # If this node itself is closer to the key than the last (furthest) node in the list,
                # we should store the value at ourselves as well
                if self._routingTables[layerID].distance(key, self.layers[layerID]) < self._routingTables[layerID].distance(key, nodes[-1].id):
                    nodes.pop()
                    try:
                        ok = self.store(key, value, originalPublisherID=originalPublisherID, age=age, expireSeconds=expireSeconds, **kwargs)
                        l.append(defer.succeed(ok))
                    except Exception as exc:
                        if _Debug:
                            traceback.print_exc()
                        l.append(defer.fail(exc))
            else:
                try:
                    ok = self.store(key, value, originalPublisherID=originalPublisherID, age=age, expireSeconds=expireSeconds, **kwargs)
                    l.append(defer.succeed(ok))
                except Exception as exc:
                    if _Debug:
                        traceback.print_exc()
                    l.append(defer.fail(exc))

            for contact in nodes:
                d = contact.store(key, value, originalPublisherID, age, expireSeconds, **kwargs)
                d.addCallback(storeSuccess, key)
                d.addErrback(storeFailed, key)
                l.append(d)

            if not collect_results:
                return nodes

            dl = defer.DeferredList(l, fireOnOneErrback=True, consumeErrors=True)
            dl.addCallback(storeRPCsCollected, nodes)
            dl.addErrback(storeRPCsFailed)
            return dl

        # Find k nodes closest to the key...
        df = self.iterativeFindNode(key, layerID=layerID, parallel_calls=parallel_calls)
        # ...and send them STORE RPCs as soon as they've been found
        df.addCallback(executeStoreRPCs)
        df.addErrback(findNodeFailed)

        if not collect_results:
            return df

        return ret

    def iterativeFindNode(self, key, layerID=0, parallel_calls=None):
        return self._iterativeFind(key, layerID=layerID, parallel_calls=parallel_calls)

    def iterativeDelete(self, key, layerID=0, parallel_calls=None):
        # Delete our own copy of the data
        if key in self._dataStores[layerID]:
            del self._dataStores[layerID][key]
        df = self._iterativeFind(key, rpc='delete', layerID=layerID, parallel_calls=parallel_calls)
        return df

    def iterativeFindValue(self, key, rpc='findValue', refresh_revision=False, layerID=0, parallel_calls=None):
        outerDf = defer.Deferred()

        def storeFailed(x):
            errmsg = getErrorMessage(x)
            if _Debug:
                print('[DHT NODE]    iterativeFindValue.storeFailed', errmsg)
            return errmsg

        def refreshRevisionSuccess(ok):
            if _Debug:
                print('[DHT NODE]    iterativeFindValue.refreshRevisionSuccess', ok)

        def refreshRevisionFailed(x):
            errmsg = getErrorMessage(x)
            if _Debug:
                print('[DHT NODE]    iterativeFindValue.refreshRevisionFailed', errmsg)
            return errmsg

        def checkResult(result):
            if _Debug:
                print('[DHT NODE]    iterativeFindValue.checkResult key=%s' % key, result)
            if isinstance(result, dict):
                if key in result:
                    latest_revision = 0
                    for v in result['values']:
                        if v[1] > latest_revision:
                            latest_revision = v[1]
                    # We have found the value; now see who was the closest contact without it...
                    if 'closestNodeNoValue' in result:
                        # ...and store the key/value pair
                        contact = result['closestNodeNoValue']
                        expireSeconds = constants.dataExpireSecondsDefaut
                        if 'expireSeconds' in result:
                            expireSeconds = result['expireSeconds']
                        if _Debug:
                            print('[DHT NODE]    republish %s to closest node with %d expire seconds' % (key, expireSeconds))
                        contact.store(key, result[key], None, 0, expireSeconds, revision=latest_revision, layerID=layerID).addErrback(storeFailed)
                    if refresh_revision:
                        # need to refresh nodes who has old version of that value
                        for v in result['values']:
                            if v[1] < latest_revision:
                                _contact = LayeredContact(encoding.to_text(v[2]), v[3][0], v[3][1], self._protocol, layerID=layerID)
                                if _Debug:
                                    print('[DHT NODE]    will refresh revision %d on %r' % (latest_revision, _contact))
                                d = _contact.store(key, result[key], None, 0, expireSeconds, revision=latest_revision, layerID=layerID)
                                d.addCallback(refreshRevisionSuccess)
                                d.addErrback(refreshRevisionFailed)
                    outerDf.callback(result)
                else:
                    # we was looking for value but did not found it
                    # Now, see if we have the value (it might seem wasteful to search on the network
                    # first, but it ensures that all values are properly propagated through the
                    # network
                    if key in self._dataStores[layerID]:
                        # Ok, we have the value locally, so use that
                        item = self._dataStores[layerID].getItem(key)
                        expireSeconds = item.get('expireSeconds', constants.dataExpireSecondsDefaut)
                        # Send this value to the closest node without it
                        if len(result['activeContacts']) > 0:
                            contact = result['activeContacts'][0]
                            if _Debug:
                                print('[DHT NODE]    refresh %s : %r with %d to %r' % (key, item['value'], expireSeconds, contact))
                            contact.store(key, item['value'], None, 0, expireSeconds, layerID=layerID).addErrback(storeFailed)
                        outerDf.callback({
                            'key': item['value'],
                            'values': [
                                (
                                    item['value'],
                                    item['revision'],
                                    self.layers[layerID],
                                    (b'127.0.0.1', self.port),
                                ),
                            ],
                            'activeContacts': result['activeContacts'],
                        })
                    else:
                        if _Debug:
                            print('[DHT NODE]    key %s does not exist in DHT' % key)
                        outerDf.callback(result)
            else:
                # The value wasn't found, but a list of contacts was returned
                # Now, see if we have the value (it might seem wasteful to search on the network
                # first, but it ensures that all values are properly propagated through the
                # network
                if key in self._dataStores[layerID]:
                    # Ok, we have the value locally, so use that
                    item = self._dataStores[layerID].getItem(key)
                    expireSeconds = item.get('expireSeconds', constants.dataExpireSecondsDefaut)
                    # Send this value to the closest node without it
                    if len(result) > 0:
                        contact = result[0]
                        if _Debug:
                            print('[DHT NODE]    refresh %s on closest node : %r with %d to %r' % (key, item['value'], expireSeconds, contact))
                        contact.store(key, item['value'], None, 0, expireSeconds, layerID=layerID).addErrback(storeFailed)
                    ret = {
                        'key': item['value'],
                        'values': [
                            (
                                item['value'],
                                item['revision'],
                                self.layers[layerID],
                                (b'127.0.0.1', self.port),
                            ),
                        ],
                        'activeContacts': result['activeContacts'],
                    }
                    if _Debug:
                        print('[DHT NODE]    key %s found in my local data store : %r' % (key, ret))
                    outerDf.callback(ret)
                else:
                    if _Debug:
                        print('[DHT NODE]    key %s does not exist in DHT' % key)
                    outerDf.callback(result)

        # Execute the search
        df = self._iterativeFind(key, rpc=rpc, layerID=layerID, parallel_calls=parallel_calls)
        df.addCallback(checkResult)
        df.addErrback(lookupFailed)
        return outerDf

    def addContact(self, contact, layerID=0):
        self._routingTables[layerID].addContact(contact)

    def removeContact(self, contactID, layerID=0):
        self._routingTables[layerID].removeContact(contactID)

    def findContact(self, contactID, layerID=0, parallel_calls=None):
        try:
            contact = self._routingTables[layerID].getContact(contactID)
            df = defer.Deferred()
            df.callback(contact)
        except ValueError:

            def parseResults(nodes):
                if contactID in nodes:
                    contact = nodes[nodes.index(contactID)]
                    return contact
                else:
                    return None

            df = self.iterativeFindNode(contactID, layerID=layerID, parallel_calls=parallel_calls)
            df.addCallback(parseResults)
        return df

    @rpcmethod
    def store(self, key, value, originalPublisherID=None, age=0, expireSeconds=constants.dataExpireSecondsDefaut, **kwargs):
        if 'layerID' in kwargs:
            layerID = kwargs['layerID']
        else:
            layerID = 0
        if layerID not in self.active_layers:
            if _Debug:
                print('[DHT NODE]    rpcmethod.store %r layerID=%d SKIP because layer is not active' % (
                    key,
                    layerID,
                ))
            raise ValueError('Layer is not active')
        if _Debug:
            print('[DHT NODE]    rpcmethod.store %r layerID=%d' % (
                key,
                layerID,
            ))
        # Get the sender's ID (if any)
        if '_rpcNodeID' in kwargs:
            rpcSenderID = kwargs['_rpcNodeID']
        else:
            rpcSenderID = None
        if originalPublisherID is None:
            if rpcSenderID is not None:
                originalPublisherID = rpcSenderID
            else:
                raise TypeError('No publisher specified, and RPC caller ID not available. Data requires an original publisher.')
        now = int(time.time())
        originallyPublished = now - age
        self._dataStores[layerID].setItem(key, value, now, originallyPublished, originalPublisherID, expireSeconds=expireSeconds, **kwargs)
        return 'OK'

    @rpcmethod
    def delete(self, key, **kwargs):
        if 'layerID' in kwargs:
            layerID = kwargs['layerID']
        else:
            layerID = 0
        if layerID not in self.active_layers:
            if _Debug:
                print('[DHT NODE]    rpcmethod.delete %r layerID=%d SKIP because layer is not active' % (
                    key,
                    layerID,
                ))
            return []
        if _Debug:
            print('[DHT NODE]    rpcmethod.delete %r layerID=%d : %r' % (key, layerID, kwargs))
        # Delete our own copy of the data (if we have one)...
        if key in self._dataStores[layerID]:
            del self._dataStores[layerID][key]
        # ...and make this RPC propagate through the network (like a FIND_VALUE for a non-existant value)
        return self.findNode(key, **kwargs)

    @rpcmethod
    def findNode(self, key, **kwargs):
        if 'layerID' in kwargs:
            layerID = kwargs['layerID']
        else:
            layerID = 0
        if layerID not in self.active_layers:
            if _Debug:
                print('[DHT NODE]    rpcmethod.findNode %r layerID=%d SKIP because layer is not active' % (
                    key,
                    layerID,
                ))
            return []
        if _Debug:
            print('[DHT NODE]    rpcmethod.findNode %r layerID=%d : %r' % (key, layerID, kwargs))
        # Get the sender's ID (if any)
        if '_rpcNodeID' in kwargs:
            rpcSenderID = kwargs['_rpcNodeID']
        else:
            rpcSenderID = None
        if layerID not in self._routingTables:
            if _Debug:
                print('[DHT NODE]    layer %d not exist' % layerID)
            return []
        contacts = self._routingTables[layerID].findCloseNodes(key, constants.k, rpcSenderID)
        contactTriples = []
        for contact in contacts:
            contactTriples.append((contact.id, encoding.to_text(contact.address), contact.port, layerID))
        if _Debug:
            print('[DHT NODE]        result is %r' % contactTriples)
        return contactTriples

    @rpcmethod
    def findValue(self, key, **kwargs):
        if 'layerID' in kwargs:
            layerID = kwargs['layerID']
        else:
            layerID = 0
        if layerID not in self.active_layers:
            if _Debug:
                print('[DHT NODE]    rpcmethod.findValue %r layerID=%d SKIP because layer is not active' % (
                    key,
                    layerID,
                ))
            return []
        if _Debug:
            print('[DHT NODE]    rpcmethod.findValue %r layerID=%r : %r' % (key, layerID, kwargs))
        if key in self._dataStores[layerID]:
            exp = None
            expireSecondsCall = getattr(self._dataStores[layerID], 'expireSeconds')
            if expireSecondsCall:
                exp = expireSecondsCall(key)
            originalPublishTimeCall = getattr(self._dataStores[layerID], 'originalPublishTime')
            published = None
            if originalPublishTimeCall:
                published = originalPublishTimeCall(key)
            if _Debug:
                print('[DHT NODE]        found key in local dataStore %r' % self._dataStores[layerID][key])
            return {
                key: self._dataStores[layerID][key],
                'expireSeconds': exp,
                'originallyPublished': published,
            }
        if _Debug:
            print('[DHT NODE]        NOT found key in local dataStore')
        return self.findNode(key, **kwargs)

    def _iterativeFind(self, key, startupShortlist=None, rpc='findNode', deep=False, layerID=0, parallel_calls=None):
        if _Debug:
            print('[DHT NODE]    _iterativeFind   layerID=%d   rpc=%r   key=%r  startupShortlist=%r routingTables=%r parallel_calls=%r' % (layerID, rpc, key, startupShortlist, self._routingTables, parallel_calls))
        if rpc != 'findNode':
            findValue = True
        else:
            findValue = False
        shortlist = []
        if startupShortlist is None:
            shortlist = self._routingTables[layerID].findCloseNodes(key, parallel_calls or constants.alpha)
            if key != self.layers[layerID]:
                # Update the "last accessed" timestamp for the appropriate k-bucket
                self._routingTables[layerID].touchKBucket(key)
            if len(shortlist) == 0:
                if _Debug:
                    print('[DHT NODE]  layerID=%d   This node doesnt know of any other nodes !!!!!' % layerID)
                # This node doesn't know of any other nodes
                fakeDf = defer.Deferred()
                fakeDf.callback([])
                return fakeDf
        else:
            # This is used during the bootstrap process; node ID's are most probably fake
            shortlist = startupShortlist
        if _Debug:
            print('[DHT NODE]   _iterativeFind  shortlist=%r' % shortlist)
        # List of active queries; len() indicates number of active probes
        # - using lists for these variables, because Python doesn't allow binding a new value to a name in an enclosing (non-global) scope
        activeProbes = []
        # List of contact IDs that have already been queried
        alreadyContacted = []
        # Probes that were active during the previous iteration
        # A list of found and known-to-be-active remote nodes
        activeContacts = []
        # This should only contain one entry; the next scheduled iteration call
        pendingIterationCalls = []
        prevClosestNode = [None]
        findValueResult = {
            'values': [],
        }
        slowNodeCount = [0]

        def extendShortlist(responseTuple):
            """ @type responseMsg: kademlia.msgtypes.ResponseMessage """
            # The "raw response" tuple contains the response message, and the originating address info
            responseMsg = responseTuple[0]
            originAddress = responseTuple[1]  # tuple: (ip adress, udp port)
            # Make sure the responding node is valid, and abort the operation if it isn't
            if _Debug:
                print('[DHT NODE]        responseTuple:', responseTuple)
            if responseMsg.nodeID in activeContacts or responseMsg.nodeID == self.layers[layerID]:
                if _Debug:
                    if responseMsg.nodeID == self.layers[layerID]:
                        print('[DHT NODE]            response from my own node')
                    else:
                        print('[DHT NODE]            response from active contact')
                return responseMsg.nodeID

            # Mark this node as active
            if responseMsg.nodeID in shortlist:
                # Get the contact information from the shortlist...
                aContact = shortlist[shortlist.index(responseMsg.nodeID)]
            else:
                # If it's not in the shortlist; we probably used a fake ID to reach it
                # - reconstruct the contact, using the real node ID this time
                aContact = LayeredContact(encoding.to_text(responseMsg.nodeID), originAddress[0], originAddress[1], self._protocol, layerID=layerID)
            activeContacts.append(aContact)
            # This makes sure "bootstrap"-nodes with "fake" IDs don't get queried twice
            if responseMsg.nodeID not in alreadyContacted:
                alreadyContacted.append(responseMsg.nodeID)
            # Now grow extend the (unverified) shortlist with the returned contacts
            result = responseMsg.response
            # TODO: some validation on the result (for guarding against attacks)
            # If we are looking for a value, first see if this result is the value
            # we are looking for before treating it as a list of contact triples
            if findValue and isinstance(result, dict):
                # We have found the value
                findValueResult[key] = result[key]
                findValueResult['values'].append((
                    result[key],
                    result.get('revision', 0),
                    responseMsg.nodeID,
                    originAddress,
                ))
                if 'expireSeconds' in result:
                    findValueResult['expireSeconds'] = result['expireSeconds']
            else:
                if findValue:
                    # We are looking for a value, and the remote node didn't have it
                    # - mark it as the closest "empty" node, if it is
                    if 'closestNodeNoValue' in findValueResult:
                        if self._routingTables[layerID].distance(key, responseMsg.nodeID) < self._routingTables[layerID].distance(key, activeContacts[0].id):
                            findValueResult['closestNodeNoValue'] = aContact
                    else:
                        findValueResult['closestNodeNoValue'] = aContact
                for contactTriple in result:
                    try:
                        testContact = LayeredContact(encoding.to_text(contactTriple[0]), contactTriple[1], contactTriple[2], self._protocol, layerID=layerID)
                    except:
                        continue
                    if testContact not in shortlist:
                        shortlist.append(testContact)
            return responseMsg.nodeID

        def removeFromShortlist(failure):
            """ @type failure: twisted.python.failure.Failure """
            failure.trap(protocol.TimeoutError)
            deadContactID = failure.getErrorMessage()
            if deadContactID in shortlist:
                if _Debug:
                    print('[DHT NODE]    removing %r' % deadContactID)
                shortlist.remove(deadContactID)
            return deadContactID

        def cancelActiveProbe(contactID):
            activeProbes.pop()
            if len(activeProbes) <= int((parallel_calls or constants.alpha)/2.0) and len(pendingIterationCalls):
                # Force the iteration
                pendingIterationCalls[0].cancel()
                del pendingIterationCalls[0]
                if _Debug:
                    print('[DHT NODE]    forcing iteration =================')
                searchIteration()

        # Send parallel, asynchronous FIND_NODE RPCs to the shortlist of contacts
        def searchIteration():
            if not self._routingTables or layerID not in self._routingTables:
                if _Debug:
                    print('[DHT NODE]    ++++++++++++++ searchIteration INTERRUPTED +++++++++++++++\n\n')
                return
            slowNodeCount[0] = len(activeProbes)
            # Sort the discovered active nodes from closest to furthest
            activeContacts.sort(key=lambda cont: self._routingTables[layerID].distance(cont.id, key))
            if _Debug:
                print('[DHT NODE]    ==> searchIteration %r' % activeContacts)
            # This makes sure a returning probe doesn't force calling this function by mistake
            while len(pendingIterationCalls):
                del pendingIterationCalls[0]
            # See if should continue the search
            if key in findValueResult and not deep:
                if _Debug:
                    print('[DHT NODE]    ++++++++++++++ DONE (findValue found) +++++++++++++++\n\n')
                findValueResult['activeContacts'] = activeContacts
                outerDf.callback(findValueResult)
                return
            if len(activeContacts) and findValue == False:
                if (len(activeContacts) >= constants.k) or (activeContacts[0] == prevClosestNode[0] and len(activeProbes) == slowNodeCount[0]):
                    # TODO: Re-send the FIND_NODEs to all of the k closest nodes not already queried
                    # Ok, we're done; either we have accumulated k active contacts or no improvement in closestNode has been noted
                    if len(activeContacts) >= constants.k:
                        if _Debug:
                            print('[DHT NODE]    ++++++++++++++ DONE (test for k active contacts) +++++++++++++++\n\n')
                    else:
                        if _Debug:
                            print('[DHT NODE]    ++++++++++++++ DONE (test for closest node) +++++++++++++++\n\n')
                    if findValue:
                        findValueResult['activeContacts'] = activeContacts
                        outerDf.callback(findValueResult)
                    else:
                        outerDf.callback(activeContacts)
                    return
            # The search continues...
            if len(activeContacts):
                prevClosestNode[0] = activeContacts[0]
            contactedNow = 0
            activeContacts.sort(key=lambda cont: self._routingTables[layerID].distance(cont.id, key))
            # Store the current shortList length before contacting other nodes
            prevShortlistLength = len(shortlist)
            for contact in shortlist:
                if contact.id not in alreadyContacted:
                    activeProbes.append(contact.id)
                    rpcMethod = getattr(contact, rpc)
                    if _Debug:
                        print('[DHT NODE] calling RPC method %r with key=%r layerID=%d at %r' % (rpc, key, layerID, contact))
                    df = rpcMethod(
                        key,
                        rawResponse=True,
                        layerID=layerID,
                    )
                    df.addCallback(extendShortlist)
                    df.addErrback(removeFromShortlist)
                    df.addCallback(cancelActiveProbe)
                    alreadyContacted.append(contact.id)
                    contactedNow += 1
                if contactedNow == (parallel_calls or constants.alpha):
                    break
            if len(activeProbes) > slowNodeCount[0] \
                    or (len(shortlist) < constants.k and len(activeContacts) < len(shortlist) and len(activeProbes) > 0):
                if _Debug:
                    print('[DHT NODE]    ----------- scheduling next call -------------')
                # Schedule the next iteration if there are any active calls (Kademlia uses loose parallelism)
                call = twisted.internet.reactor.callLater(constants.iterativeLookupDelay, searchIteration)  # IGNORE:E1101  @UndefinedVariable
                pendingIterationCalls.append(call)
            # Check for a quick contact response that made an update to the shortList
            elif prevShortlistLength < len(shortlist):
                # Ensure that the closest contacts are taken from the updated shortList
                searchIteration()
            else:
                if _Debug:
                    print('[DHT NODE]    ++++++++++++++ DONE (logically) +++++++++++++\n\n')
                # If no probes were sent, there will not be any improvement, so we're done
                if findValue:
                    findValueResult['activeContacts'] = activeContacts
                    outerDf.callback(findValueResult)
                else:
                    outerDf.callback(activeContacts)

        outerDf = defer.Deferred()
        # Start the iterations
        searchIteration()
        return outerDf

    def _persistState(self, *args, **kwargs):
        layerID = kwargs.get('layerID', 0)
        self._joinDeferreds.pop(layerID, None)
        closestNodes = list(self.findNode(self.layers[layerID], **kwargs))
        state = {
            'id': self.layers[layerID],
            'closestNodes': closestNodes,
            'key': 'nodeState',
            'type': 'skip_validation',
        }
        if _Debug:
            print('[DHT NODE]    _persistState  layerID=%d id=%r closestNodes=%r state=%r' % (
                layerID,
                self.layers[layerID],
                closestNodes,
                state,
            ))
        json_value = json.dumps(state)
        now = int(time.time())
        self._dataStores[layerID].setItem(self.nodeStateKey, json_value, now, now, self.layers[layerID])
        return args

    def _joinNetworkFailed(self, err, **kwargs):
        layerID = kwargs.get('layerID', 0)
        self._joinDeferreds.pop(layerID, None)
        if _Debug:
            print('[DHT NODE]    failed joining DHT network layerID=%d ' % layerID, err)

    def _refreshNode(self, layerID=0):
        """
        Periodically called to perform k-bucket refreshes and data
        replication/republishing as necessary.
        """
        # if self._counter:
        #     self._counter('_refreshNode')
        if layerID not in self.layers:
            return
        df = self._refreshRoutingTable(layerID=layerID)
        df.addCallback(self._republishData, layerID=layerID)
        df.addCallback(self._scheduleNextNodeRefresh, layerID=layerID)

    def _refreshRoutingTable(self, layerID=0, parallel_calls=None):
        nodeIDs = self._routingTables[layerID].getRefreshList(0, False)
        if _Debug:
            print('[DHT NODE]    _refreshRoutingTable', nodeIDs)

        outerDf = defer.Deferred()

        def searchFailed(err):
            if _Debug:
                print('[DHT NODE]    searchFailed', err)

        def searchForNextNodeID(dfResult=None):
            if len(nodeIDs) > 0:
                searchID = nodeIDs.pop()
                df = self.iterativeFindNode(searchID, layerID=layerID, parallel_calls=parallel_calls)
                df.addCallback(searchForNextNodeID)
                df.addErrback(searchFailed)
            else:
                # If this is reached, we have finished refreshing the routing table
                outerDf.callback(None)

        # Start the refreshing cycle
        searchForNextNodeID()
        return outerDf

    def _republishData(self, *args, **kwargs):
        df = twisted.internet.threads.deferToThread(self._threadedRepublishData, *args, **kwargs)
        return df

    def _scheduleNextNodeRefresh(self, *args, **kwargs):
        if _Debug:
            print('[DHT NODE] will refresh layer %d in %d seconds' % (
                kwargs['layerID'],
                constants.checkRefreshInterval,
            ))
        self.refreshers[kwargs['layerID']] = twisted.internet.reactor.callLater(  # @UndefinedVariable
            constants.checkRefreshInterval,
            self._refreshNode,
            layerID=kwargs['layerID'],
        )  # @UndefinedVariable

    def _threadedRepublishData(self, *args, **kwargs):
        """
        Republishes and expires any stored data (i.e. stored C{(key, value
        pairs)} that need to be republished/expired.

        This method should run in a deferred thread
        """
        layerID = kwargs['layerID']
        if _Debug:
            print('[DHT NODE]    republishData called, node: %r' % self.layers[layerID])
        expiredKeys = []
        for key in self._dataStores[layerID].keys():
            if _Debug:
                print('[DHT NODE]        %r' % key)
            # Filter internal variables stored in the datastore
            if key == 'nodeState':
                continue

            now = int(time.time())
            itemData = self._dataStores[layerID].getItem(key)
            originallyPublished = itemData['originallyPublished']
            originalPublisherID = itemData['originalPublisherID']
            lastPublished = itemData['lastPublished']
            expireSeconds = itemData['expireSeconds']
            age = now - originallyPublished
            if originalPublisherID == self.layers[layerID]:
                # This node is the original publisher; it has to republish
                # the data before it expires (24 hours in basic Kademlia)
                if age >= constants.dataExpireTimeout:
                    twisted.internet.reactor.callFromThread(  # @UndefinedVariable
                        self.iterativeStore,
                        key=key,
                        value=itemData['value'],
                        expireSeconds=expireSeconds,
                        layerID=layerID,
                    )
            else:
                # This node needs to replicate the data at set intervals,
                # until it expires, without changing the metadata associated with it
                # First, check if the data has expired
                if age >= constants.dataExpireTimeout:
                    # This key/value pair has expired (and it has not been republished by the original publishing node
                    # - remove it
                    expiredKeys.append(key)
                elif now - lastPublished >= constants.replicateInterval:
                    # ...data has not yet expired, and we need to replicate it
                    twisted.internet.reactor.callFromThread(  # @UndefinedVariable
                        self.iterativeStore,
                        key=key,
                        value=self._dataStore[key],
                        originalPublisherID=originalPublisherID,
                        age=age,
                        expireSeconds=expireSeconds,
                        layerID=layerID,
                    )
        for key in expiredKeys:
            del self._dataStores[layerID][key]

if __name__ == '__main__':
    import sys
    if len(sys.argv) < 2:
        print('Usage:\n%s UDP_PORT  [KNOWN_NODE_IP  KNOWN_NODE_PORT]' % sys.argv[0])
        print('or:\n%s UDP_PORT  [FILE_WITH_KNOWN_NODES]' % sys.argv[0])
        print('\nIf a file is specified, it should containg one IP address and UDP port\nper line, seperated by a space.')
        sys.exit(1)
    try:
        usePort = int(sys.argv[1])
    except ValueError:
        print('\nUDP_PORT must be an integer value.\n')
        print('Usage:\n%s UDP_PORT  [KNOWN_NODE_IP  KNOWN_NODE_PORT]' % sys.argv[0])
        print('or:\n%s UDP_PORT  [FILE_WITH_KNOWN_NODES]' % sys.argv[0])
        print('\nIf a file is specified, it should contain one IP address and UDP port\nper line, seperated by a space.')
        sys.exit(1)

    if len(sys.argv) == 4:
        knownNodes = [(sys.argv[2], int(sys.argv[3]))]
    elif len(sys.argv) == 3:
        knownNodes = []
        f = open(sys.argv[2], 'r')
        lines = f.readlines()
        f.close()
        for line in lines:
            ipAddress, udpPort = line.split()
            knownNodes.append((ipAddress, int(udpPort)))
    else:
        knownNodes = None

    node = Node(udpPort=usePort)
    node.joinNetwork(knownNodes)
    twisted.internet.reactor.run()  # @UndefinedVariable



import psutil


def proccess_presence(process_name):
    for process in psutil.pids():
        try:
            p = psutil.Process(process)  # The pid of desired process

            # print(p.name()) # If the name is "python.exe" is called by python
            # print(p.cmdline()) # Is the command line this process has been called with
            if process_name in str(p.cmdline()):
                return True
        except:
            pass

    return False


if __name__ == '__main__':
    print(proccess_presence('node.py'))



#!/usr/bin/python
# coins_io.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (coins_io.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: coins_io

Every "contract" store a list of "coins", they form a single "chain" in the global DB
This is similar to well-known "blockchain" technology where every block is linked to another by hash.

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------

import json

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import sys
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import utime

from bitdust.crypt import key

from bitdust.userid import my_id

#------------------------------------------------------------------------------


def storage_contract_open(customer_idurl, duration, amount, price=1.0, trustee=None):
    """
    + type of resources: storage space (can be also : cpu, traffic, etc.)
    + amount: in megabytes
    + duration: seconds
    + customer: idurl of consumer
    + supplier: idurl of provider
    + trustee: idurl of trusted supporter
    + started: seconds since epoch
    + price: 1.0 by default
    """
    return {
        'miner': {
            'prev': '',
        },
        'payload': {
            'type': 'storage',
            'amount': amount,
            'duration': duration,
            'customer': customer_idurl,
            'supplier': my_id.getIDURL(),
            'trustee': trustee,
            'started': utime.utcnow_to_sec1970(),
            'price': price,
        },
    }


def storage_contract_accept(prev_coin_json):
    return {
        'miner': {
            # populate previous hash from existing coin
            'prev': prev_coin_json['miner']['hash'],
        },
        'payload': {
            'accepted': utime.utcnow_to_sec1970(),
        },
    }


def storage_contract_continue(prev_coin_json, duration):
    return {
        'miner': {
            # populate previous hash from existing coin
            'prev': prev_coin_json['miner']['hash'],
        },
        'payload': {
            'extended': utime.utcnow_to_sec1970(),
            'duration': duration,
        },
    }


def add_signature(coin_json, role):
    _coin = coin_json.copy()
    _coin[role] = {
        'idurl': my_id.getIDURL().to_bin(),
        'pubkey': key.MyPublicKey(),
    }
    coin_hash = get_coin_hash(_coin)
    _coin[role]['signature'] = key.Sign(coin_hash)
    return _coin


def verify_signature(coin_json, role):
    signature = coin_json[role]['signature']
    coin_json[role]['signature'] = ''
    # role_data = coin_json[role].copy()
    # signature = role_data.pop('signature')
    # _coin = dict(coin_json).c
    coin_hash = get_coin_hash(coin_json)
    coin_json[role]['signature'] = signature
    return key.VerifySignature(coin_json[role]['pubkey'], coin_hash, signature)


def set_prev_hash(coin_json, prev_hash):
    if 'miner' not in coin_json:
        coin_json['miner'] = {}
    coin_json['miner']['prev'] = prev_hash
    return coin_json


# def signed_coin(coin_json):
#     scoin = {
#         "creator": my_id.getIDURL(),
#         "pubkey": key.MyPublicKey(),
#     }
#     scoin.update(coin_json)
#     chash = get_coin_hash(scoin)
#     signature = key.Sign(chash)
#     scoin['signature'] = signature
#     return scoin


def coin_to_string(coin_json):
    return json.dumps(coin_json, sort_keys=True)


def coins_to_string(coins_list):
    return json.dumps(coins_list, sort_keys=True)


def string_to_coin(s):
    return json.loads(s)


def get_coin_hash(coin_json):
    coin_hashbase = coin_to_string(coin_json)
    return key.Hash(coin_hashbase, hexdigest=True)


def get_coin_base(coin_json):
    return coin_json


#     bcoin = coin_json.copy()
#     bcoin.pop('creator')
#     bcoin.pop('signature')
#     bcoin.pop('pubkey')
#     return bcoin


def bought_storage(partner):
    pass


#------------------------------------------------------------------------------


def read_query_from_packet(newpacket):
    try:
        query_j = json.loads(newpacket.Payload)
    except:
        lg.exc()
        return None
    # TODO: verify input query fields
    return query_j


def read_coins_from_packet(newpacket):
    try:
        coins_list = json.loads(newpacket.Payload)
    except:
        lg.exc()
        return None
    # TODO: verify all input coins here
    return coins_list


#------------------------------------------------------------------------------


def validate_coin(coin_json):
    # TODO: validate sub-fields, hashes, query on DB, etc.
    # if 'miner' not in coin_json:
    #     return False
    if 'creator' not in coin_json:
        return False
    if 'signature' not in coin_json['creator']:
        return False
    # if 'signer' not in coin_json:
    #     return False
    if 'payload' not in coin_json:
        return False
    return True


def verify_coin(coin_json, role):
    # TODO: check coin was really mined
    # if coin_json['hash'] != calculated_hash:
    #     return False
    return True


#------------------------------------------------------------------------------

#------------------------------------------------------------------------------


def _test():
    pass


if __name__ == '__main__':
    lg.set_debug_level(20)
    _test()



#!/usr/bin/env python
# ppworker.py
#
# Parallel Python Software: http://www.parallelpython.com
# Copyright (c) 2005-2009, Vitalii Vanovschi
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the author nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
"""
Parallel Python Software, PP Worker.

http://www.parallelpython.com - updates, documentation, examples and support
forums
"""

from __future__ import absolute_import
from __future__ import print_function
import six
from io import BytesIO

copyright = 'Copyright (c) 2005-2009 Vitalii Vanovschi. All rights reserved'
version = '1.5.7'

_Debug = False

import sys
import os
import json
import traceback

from . import pptransport


def import_module(name):
    mod = __import__(name)
    components = name.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def preprocess(msg):
    try:
        fname, fsources, imports = json.loads(msg)['v']
        fobjs = [compile(fsource, '<string>', 'exec') for fsource in fsources]
        for module in imports:
            try:
                globals()[module.split('.')[0]] = __import__(module)
            except:
                if _Debug:
                    open('/tmp/raid.log', 'a').write(u'%s\n' % traceback.format_exc())
                # print("An error has occured during the module import")
                sys.excepthook(*sys.exc_info())
        return fname, fobjs
    except Exception as exc:
        if _Debug:
            open('/tmp/raid.log', 'a').write(u'%s\n%s\n' % (traceback.format_exc(), msg))


class _WorkerProcess(object):
    def __init__(self):
        self.hashmap = {}
        self.e = sys.__stderr__
        self.sout = BytesIO()
        origsout = sys.stdout
        sys.stdout = self.sout
        sys.stderr = self.sout
        sin = sys.stdin
        self.t = pptransport.PipeTransport(sin, origsout)
        self.t.send(six.text_type(os.getpid()))

    def run(self):
        try:
            # execution cycle
            while True:

                __fname, __fobjs = self.t.receive(preprocess)

                __sargs = self.t.receive()

                for __fobj in __fobjs:
                    try:
                        eval(__fobj)
                        globals().update(locals())
                    except:
                        # print("An error has occured during the " + \
                        #       "function import")
                        sys.excepthook(*sys.exc_info())

                __args = json.loads(__sargs)['v']

                __f = locals()[__fname]
                try:
                    __result = __f(*__args)
                except:
                    # print("An error has occured during the function execution")
                    sys.excepthook(*sys.exc_info())
                    __result = None

                __sresult = json.dumps({
                    'v': (__result, self.sout.getvalue().decode('latin1')),
                })

                self.t.send(__sresult)
                self.sout.truncate(0)
        except:
            # print("Fatal error has occured during the function execution")
            if _Debug:
                open('/tmp/raid.log', 'a').write(u'%s\n' % traceback.format_exc())
            sys.excepthook(*sys.exc_info())
            __result = None
            __sresult = json.dumps({
                'v': (__result, self.sout.getvalue().decode('latin1')),
            })

            self.t.send(__sresult)


if __name__ == '__main__':
    sys.path.append(os.path.dirname(__file__))
    wp = _WorkerProcess()
    wp.run()

# Parallel Python Software: http://www.parallelpython.com



#!/usr/bin/env python
# proxy_sender.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (proxy_sender.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: proxy_sender.

.. role:: red

BitDust proxy_sender() Automat

.. raw:: html

    <a href="proxy_sender.png" target="_blank">
    <img src="proxy_sender.png" style="max-width:100%;">
    </a>

EVENTS:
    * :red:`init`
    * :red:`proxy_receiver.state`
    * :red:`relay-ack`
    * :red:`relay-failed`
    * :red:`relay-out`
    * :red:`retry`
    * :red:`retry-cache-failed`
    * :red:`retry-failed`
    * :red:`retry-send-failed`
    * :red:`shutdown`
    * :red:`start`
    * :red:`stop`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 16

_PacketLogFileEnabled = False

#------------------------------------------------------------------------------

from twisted.internet.defer import Deferred
from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.automats import automat

from bitdust.logs import lg

from bitdust.lib import nameurl
from bitdust.lib import serialization
from bitdust.lib import net_misc

from bitdust.main import config
from bitdust.main import settings

from bitdust.crypt import encrypted
from bitdust.crypt import key
from bitdust.crypt import signed

from bitdust.services import driver

from bitdust.contacts import identitycache

from bitdust.p2p import commands
from bitdust.p2p import network_connector

from bitdust.transport import callback
from bitdust.transport import packet_out

from bitdust.transport.proxy import proxy_receiver

from bitdust.userid import id_url
from bitdust.userid import global_id
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_ProxySender = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with proxy_sender machine.
    """
    global _ProxySender
    if event is None and not args:
        return _ProxySender
    if _ProxySender is None:
        # set automat name and starting state here
        _ProxySender = ProxySender(
            name='proxy_sender',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=False,
            log_transitions=_Debug,
        )
    if event is not None:
        _ProxySender.automat(event, *args, **kwargs)
    return _ProxySender


#------------------------------------------------------------------------------


class ProxySender(automat.Automat):
    """
    This class implements all the functionality of the ``proxy_sender()`` state
    machine.
    """
    def init(self, **kwargs):
        global _PacketLogFileEnabled
        _PacketLogFileEnabled = config.conf().getBool('logs/packet-enabled')
        self._sending_enabled = settings.enablePROXYsending()

    def to_json(self):
        j = super().to_json()
        j.update(
            {
                'proto': proxy_receiver.GetRouterProtoHost()[0] if proxy_receiver.GetRouterProtoHost() else '',
                'host': net_misc.pack_address_text(proxy_receiver.GetRouterProtoHost()[1]) if proxy_receiver.GetRouterProtoHost() else '',
                'idurl': proxy_receiver.GetRouterIDURL(),
                'bytes_received': 0,
                'bytes_sent': self.traffic_out,
            }
        )
        return j

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python
        <http://code.google.com/p/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'STOPPED'
                self.doInit(*args, **kwargs)
        #---STOPPED---
        elif self.state == 'STOPPED':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'start' and proxy_receiver.A().state != 'LISTEN' and self.isSendingEnabled(*args, **kwargs):
                self.state = 'ROUTER?'
                self.doStartFilterOutgoingTraffic(*args, **kwargs)
            elif ((event == 'proxy_receiver.state' and args[0] == 'LISTEN') or (event == 'start' and proxy_receiver.A().state == 'LISTEN')) and self.isSendingEnabled(*args, **kwargs):
                self.state = 'REDIRECTING'
                self.doStartFilterOutgoingTraffic(*args, **kwargs)
        #---ROUTER?---
        elif self.state == 'ROUTER?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doStopFilterOutgoingTraffic(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif (event == 'proxy_receiver.state' and args[0] == 'LISTEN'):
                self.state = 'REDIRECTING'
                self.doSendAllPendingPackets(*args, **kwargs)
            elif (event == 'proxy_receiver.state' and args[0] == 'OFFLINE'):
                self.state = 'STOPPED'
                self.doStopFilterOutgoingTraffic(*args, **kwargs)
        #---REDIRECTING---
        elif self.state == 'REDIRECTING':
            if event == 'stop':
                self.state = 'STOPPED'
                self.doStopFilterOutgoingTraffic(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doStopFilterOutgoingTraffic(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'relay-failed':
                self.doRetryCancelPacket(*args, **kwargs)
            elif event == 'retry' or event == 'retry-cache-failed' or event == 'retry-send-failed' or event == 'retry-failed':
                self.doReportRetry(event, *args, **kwargs)
            elif event == 'relay-ack':
                self.doCleanPacket(*args, **kwargs)
            elif event == 'relay-out':
                self.doCountTraffic(*args, **kwargs)
            elif (event == 'proxy_receiver.state' and args[0] != 'LISTEN'):
                self.state = 'ROUTER?'
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def isSendingEnabled(self, *args, **kwargs):
        """
        Condition method.
        """
        return self._sending_enabled

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.traffic_out = 0
        self.pending_packets = []
        self.pending_ping_packets = []
        self.max_pending_packets = 100  # TODO: read from settings
        self.packets_retries = {}
        self.sent_packets = {}

    def doStartFilterOutgoingTraffic(self, *args, **kwargs):
        """
        Action method.
        """
        callback.insert_outbox_filter_callback(0, self._on_first_outbox_packet)

    def doStopFilterOutgoingTraffic(self, *args, **kwargs):
        """
        Action method.
        """
        callback.remove_outbox_filter_callback(self._on_first_outbox_packet)

    def doCountTraffic(self, *args, **kwargs):
        """
        Action method.
        """
        _, newpacket, _ = args[0]
        self.traffic_out += len(newpacket.Payload)

    def doCleanPacket(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_clean_sent_packet(args[0])

    def doRetryCancelPacket(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_retry_one_time(args[0])

    def doReportRetry(self, event, *args, **kwargs):
        """
        Action method.
        """
        if _PacketLogFileEnabled:
            label = event.upper().replace('-', ' ')
            fail_info = args[0]
            lg.out(
                0,
                '\033[0;49;36m%s %s(%s) from %s to %s %s\033[0m' % (label, fail_info['command'], fail_info['packet_id'], global_id.UrlToGlobalID(fail_info['from']), global_id.UrlToGlobalID(fail_info['to']), fail_info['error']),
                log_name='packet',
                showtime=True,
            )

    def doSendAllPendingPackets(self, *args, **kwargs):
        """
        Action method.
        """
        def _do_send():
            while len(self.pending_packets):
                outpacket, callbacks, wide, response_timeout, keep_alive, pending_result = self.pending_packets.pop(0)
                if _Debug:
                    lg.out(_DebugLevel, 'proxy_sender.doSendAllPendingPackets populate one more item, %d more in the queue' % (len(self.pending_packets)))
                result_packet = self._on_first_outbox_packet(outpacket, wide, callbacks, response_timeout=response_timeout, keep_alive=keep_alive)
                if not isinstance(result_packet, packet_out.PacketOut):
                    lg.warn('failed sending pending packet %s, skip all pending packets' % outpacket)
                    self.pending_packets = []
                    break
                pending_result.callback(result_packet)

        reactor.callLater(0, _do_send)  # @UndefinedVariable

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        global _PacketLogFileEnabled
        _PacketLogFileEnabled = False
        self.traffic_out = 0
        self.pending_packets = []
        self.pending_ping_packets = []
        self.max_pending_packets = 0
        self.packets_retries.clear()
        self.sent_packets.clear()
        self.destroy()
        global _ProxySender
        del _ProxySender
        _ProxySender = None

    def _do_add_pending_packet(self, outpacket, callbacks, wide, response_timeout, keep_alive):
        if len(self.pending_packets) > self.max_pending_packets:
            if _Debug:
                lg.warn('pending packets queue is full, skip sending outgoing packet')
            return None
        pending_result = Deferred()
        self.pending_packets.append((outpacket, callbacks, wide, response_timeout, keep_alive, pending_result))
        if _Debug:
            lg.out(_DebugLevel, 'proxy_sender._do_add_pending_packet %s' % outpacket)
        return pending_result

    def _do_send_packet_to_router(self, outpacket, callbacks, wide, response_timeout, keep_alive, is_retry=False):
        router_idurl = proxy_receiver.GetRouterIDURL()
        router_identity_obj = proxy_receiver.GetRouterIdentity()
        router_proto_host = proxy_receiver.GetRouterProtoHost()
        router_proto, router_host = router_proto_host
        publickey = router_identity_obj.publickey
        my_original_identity_src = proxy_receiver.ReadMyOriginalIdentitySource()
        if not router_idurl or not router_identity_obj or not router_proto_host or not my_original_identity_src:
            if _Debug:
                lg.out(_DebugLevel, 'proxy_sender._do_send_packet_to_router SKIP because router not ready yet')
            return self._do_add_pending_packet(outpacket, callbacks, wide, response_timeout, keep_alive)
        if outpacket.RemoteID.to_bin() == router_idurl.to_bin():
            if _Debug:
                lg.out(_DebugLevel, 'proxy_sender._do_send_packet_to_router SKIP, packet addressed to router and must be sent in a usual way')
            return None
        try:
            raw_data = outpacket.Serialize()
        except:
            lg.exc('failed to Serialize %s' % outpacket)
            return None
        # see proxy_router.ProxyRouter : doForwardOutboxPacket() for receiving part
        json_payload = {
            'f': my_id.getIDURL().to_bin(),  # from
            't': outpacket.RemoteID.to_bin(),  # to
            'p': raw_data,  # payload
            'w': wide,  # wide
            'i': response_timeout,
            'a': keep_alive,
            'r': is_retry,
        }
        if not json_payload['t']:
            raise ValueError('receiver idurl was not set')
        raw_bytes = serialization.DictToBytes(json_payload)
        block = encrypted.Block(
            CreatorID=my_id.getIDURL(),
            BackupID='routed outgoing data',
            BlockNumber=0,
            SessionKey=key.NewSessionKey(session_key_type=key.SessionKeyType()),
            SessionKeyType=key.SessionKeyType(),
            LastBlock=True,
            Data=raw_bytes,
            EncryptKey=lambda inp: key.EncryptOpenSSHPublicKey(publickey, inp),
        )
        block_encrypted = block.Serialize()
        newpacket = signed.Packet(
            Command=commands.RelayOut(),
            OwnerID=outpacket.OwnerID,
            CreatorID=my_id.getIDURL(),
            PacketID=outpacket.PacketID,
            Payload=block_encrypted,
            RemoteID=router_idurl,
        )
        if response_timeout is not None:
            # must give some extra time for the proxy re-routing
            response_timeout += 10.0
        routed_packet = packet_out.create(
            outpacket=outpacket,
            wide=False,
            callbacks={},
            route={
                'packet': newpacket,  # pointing "newpacket" to router node
                'proto': router_proto,
                'host': router_host,
                'remoteid': router_idurl,
                'description': 'RelayOut_%s[%s]_%s' % (outpacket.Command, outpacket.PacketID, nameurl.GetName(router_idurl)),
            },
            response_timeout=response_timeout,
            keep_alive=True,
        )
        for command, cb_list in callbacks.items():
            if isinstance(cb_list, list):
                for cb in cb_list:
                    routed_packet.set_callback(command, cb)
            else:
                routed_packet.set_callback(command, cb_list)
        if not is_retry:
            _key = (outpacket.Command, outpacket.PacketID, outpacket.RemoteID.to_bin())
            self.sent_packets[_key] = (routed_packet, outpacket)
        self.event('relay-out', (outpacket, newpacket, routed_packet))
        if _Debug:
            lg.out(_DebugLevel, '>>>Relay-OUT %s sent to %s://%s with %d bytes, timeout=%r' % (str(outpacket), router_proto, router_host, len(block_encrypted), response_timeout))
        if _PacketLogFileEnabled:
            lg.out(
                0, '\033[0;49;36mRELAY OUT %s(%s) with %s bytes from %s to %s via %s\033[0m' %
                (outpacket.Command, outpacket.PacketID, len(raw_bytes), global_id.UrlToGlobalID(outpacket.CreatorID), global_id.UrlToGlobalID(outpacket.RemoteID), global_id.UrlToGlobalID(router_idurl)), log_name='packet', showtime=True
            )
        del raw_bytes
        del block
        del newpacket
        del outpacket
        del router_identity_obj
        del router_idurl
        del router_proto_host
        return routed_packet

    def _do_cancel_outbox_packets(self, fail_info):
        to_idurl = id_url.field(fail_info['to'])
        from_idurl = id_url.field(fail_info['from'])
        for p in packet_out.search_by_packet_id(fail_info['packet_id']):
            if p.outpacket.Command == fail_info['command']:
                if id_url.to_bin(to_idurl) == p.outpacket.RemoteID.to_bin():
                    if p.outpacket.CreatorID.to_bin() == id_url.to_bin(from_idurl) or p.outpacket.OwnerID.to_bin() == id_url.to_bin(from_idurl):
                        if _Debug:
                            lg.dbg(_DebugLevel, 'about to cancel %r because sending via proxy transport is failed' % p)
                        p.automat('cancel')

    def _do_retry_one_time(self, fail_info):
        to_idurl = id_url.field(fail_info['to']).to_bin()
        from_idurl = id_url.field(fail_info['from']).to_bin()
        _key = (fail_info['command'], fail_info['packet_id'], from_idurl, to_idurl)
        current_retries = self.packets_retries.get(_key, 0)
        if _Debug:
            lg.args(_DebugLevel, key=_key, retries=current_retries)
        if fail_info.get('error') != 'route already closed':
            if _Debug:
                lg.dbg(_DebugLevel, 'failed sending routed packet : %r' % fail_info)
            self._do_clean_sent_packet(fail_info)
            self._do_cancel_outbox_packets(fail_info)
            self.packets_retries.pop(_key, None)
            return
        if current_retries >= 1:
            if _Debug:
                lg.dbg(_DebugLevel, 'failed sending routed packet after few attempts : %r' % fail_info)
            self.automat('retry-failed', fail_info)
            self._do_clean_sent_packet(fail_info)
            self._do_cancel_outbox_packets(fail_info)
            self.packets_retries.pop(_key, None)
            return
        self.packets_retries[_key] = current_retries + 1
        d = identitycache.immediatelyCaching(fail_info['to'])
        d.addCallback(self._on_cache_retry_success, fail_info)
        d.addErrback(self._on_cache_retry_failed, fail_info)

    def _do_clean_sent_packet(self, info):
        to_idurl = id_url.to_bin(info['to'])
        to_remove = []
        for _key in self.sent_packets.keys():
            routed_packet, outpacket = self.sent_packets.get(_key, (
                None,
                None,
            ))
            if not outpacket:
                if _Debug:
                    lg.dbg(_DebugLevel, 'found empty outpacket : %r' % routed_packet)
                to_remove.append(_key)
                continue
            if outpacket.Command != info['command']:
                continue
            if outpacket.PacketID != info['packet_id']:
                continue
            if outpacket.RemoteID.to_bin() != to_idurl:
                continue
            to_remove.append(_key)
        for _key in to_remove:
            routed_packet, outpacket = self.sent_packets.pop(_key, (
                None,
                None,
            ))

    def _on_cache_retry_success(self, xmlsrc, fail_info):
        if _Debug:
            lg.args(_DebugLevel, sent_packets=len(self.sent_packets), fail_info=fail_info)
        to_idurl = id_url.to_bin(fail_info['to'])
        for _key in self.sent_packets.keys():
            routed_packet, outpacket = self.sent_packets.get(_key, (
                None,
                None,
            ))
            if not outpacket:
                if _Debug:
                    lg.dbg(_DebugLevel, 'found empty outpacket : %r' % routed_packet)
                continue
            if outpacket.Command != fail_info['command']:
                continue
            if outpacket.PacketID != fail_info['packet_id']:
                continue
            if outpacket.RemoteID.to_bin() != to_idurl:
                continue
            routed_retry_packet = self._do_send_packet_to_router(
                outpacket=outpacket,
                callbacks=routed_packet.callbacks,
                wide=fail_info.get('wide', False),
                keep_alive=fail_info.get('keep_alive', False),
                response_timeout=fail_info.get('response_timeout', None),
                is_retry=True,
            )
            if not routed_retry_packet:
                self.automat('retry-send-failed', fail_info)
            else:
                self.sent_packets[_key] = (
                    routed_retry_packet,
                    outpacket,
                )
                self.automat('retry', fail_info)
            del routed_packet
        return None

    def _on_cache_retry_failed(self, err, fail_info):
        if _Debug:
            lg.args(_DebugLevel, err=err, fail_info=fail_info)
        self.automat('retry-cache-failed', fail_info)
        self._do_cancel_outbox_packets(fail_info)
        return None

    def _on_first_outbox_packet(self, outpacket, wide, callbacks, target=None, route=None, response_timeout=None, keep_alive=True):
        """
        Will be called first for every outgoing packet.
        Must return `None` if that packet should be sent in a normal way.
        Otherwise will create another "routed" packet instead and return it.
        """
        if not driver.is_on('service_proxy_transport'):
            if _Debug:
                lg.out(_DebugLevel, 'proxy_sender._on_first_outbox_packet SKIP sending %r because service_proxy_transport is not started yet' % outpacket)
            return None
        if not proxy_receiver.A():
            if _Debug:
                lg.out(_DebugLevel, 'proxy_sender._on_first_outbox_packet SKIP sending %r because proxy_receiver() not exist' % outpacket)
            return None
        if outpacket.Command == commands.Identity() and outpacket.CreatorID == my_id.getIDURL():
            if proxy_receiver.GetPossibleRouterIDURL() and proxy_receiver.GetPossibleRouterIDURL().to_bin() == outpacket.RemoteID.to_bin():
                if network_connector.A().state == 'DISCONNECTED':
                    if _Debug:
                        lg.out(_DebugLevel, 'proxy_sender._on_first_outbox_packet SKIP sending %r because network_connector() is DISCONNECTED' % outpacket)
                    return None
                if network_connector.A().state == 'CONNECTED':
                    lg.warn('sending %r to "possible" proxy router %r' % (outpacket, proxy_receiver.GetPossibleRouterIDURL()))
                    pkt_out = packet_out.create(outpacket, wide, callbacks, target, route, response_timeout, keep_alive)
                    return pkt_out
                if _Debug:
                    lg.out(_DebugLevel, 'proxy_sender._on_first_outbox_packet SKIP sending %r, network_connector() have transition state' % outpacket)
                return None
        if proxy_receiver.A().state != 'LISTEN':
            if _Debug:
                lg.out(_DebugLevel, 'proxy_sender._on_first_outbox_packet DELLAYED %r because proxy_receiver state is not LISTEN yet' % outpacket)
            return self._do_add_pending_packet(outpacket, callbacks, wide, response_timeout, keep_alive)
        return self._do_send_packet_to_router(
            outpacket=outpacket,
            callbacks=callbacks,
            wide=wide,
            keep_alive=keep_alive,
            response_timeout=response_timeout,
        )

    def _on_network_connector_state_changed(self, oldstate, newstate, event_string, *args, **kwargs):
        if newstate == 'CONNECTED' and oldstate != newstate:
            if _Debug:
                lg.out(_DebugLevel, 'proxy_sender._on_network_connector_state_changed will send %d pending "ping" packets' % len(self.pending_ping_packets))
            while len(self.pending_ping_packets):
                outpacket, wide, callbacks, target, route, response_timeout, keep_alive, pending_result = self.pending_ping_packets.pop(0)
                if _Debug:
                    lg.out(_DebugLevel, 'proxy_sender._on_network_connector_state_changed populate one more item, %d more in the queue' % (len(self.pending_packets)))
                result_packet = self._on_first_outbox_packet(outpacket, wide, callbacks, target, route, response_timeout, keep_alive)
                if not isinstance(result_packet, packet_out.PacketOut):
                    lg.warn('failed sending pending packet %s, skip all pending packets' % outpacket)
                    self.pending_ping_packets = []
                    break
                pending_result.callback(result_packet)


#------------------------------------------------------------------------------


def main():
    reactor.callWhenRunning(A, 'init')  # @UndefinedVariable
    reactor.run()  # @UndefinedVariable


#------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



import essentials, options, connections, socks

config = options.Get()
config.read()
node_ip = config.node_ip
port = config.port

key, public_key_readable, private_key_readable, encrypted, unlocked, public_key_b64encoded, address = essentials.keys_load('privkey.der', 'pubkey.der')

s = socks.socksocket()
s.settimeout(10)
s.connect((node_ip, int(port)))

connections.send(s, 'balanceget', 10)
connections.send(s, address, 10)
balanceget_result = connections.receive(s, 10)
print('Address balance: {}'.format(balanceget_result[0]))
print('Address credit: {}'.format(balanceget_result[1]))
print('Address debit: {}'.format(balanceget_result[2]))
print('Address fees: {}'.format(balanceget_result[3]))
print('Address rewards: {}'.format(balanceget_result[4]))
print('Address balance without mempool: {}'.format(balanceget_result[5]))
# get balance



#!/usr/bin/env python
#__init__.py
#
# Copyright (C) 2007-2008 Francois Aucamp, Meraka Institute, CSIR
# See AUTHORS for all authors and contact information.
#
# License: GNU Lesser General Public License, version 3 or later; see COPYING
#          included in this archive for details.
#
# This library is free software, distributed under the terms of
# the GNU Lesser General Public License Version 3, or any later version.
# See the COPYING file included in this archive
#
# The docstrings in this module contain epytext markup; API documentation
# may be created by processing this file with epydoc: http://epydoc.sf.net
"""
Entangled DHT and distributed tuple space.

The distributed hash table (DHT) used by Entangled is based on Kademlia,
and be accessed by the C{entangled.kademlia package}, or by simply
instantiating/subclassing the exposed C{KademliaNode} in the main C{entangled}
package.

On top of this Kademlia node Entangled provides some extra functionality
in the form of a "C{DELETE}" RPC and keyword-based search operations; these
functions are accessible via the C{EntangledNode} class in the main
C{entangled} package.

The Entangled distributed tuple space is exposed as the
C{DistributedTupleSpacePeer} class, accessible via the main C{entangled}
package or its C{dtuple} module.
"""

from __future__ import absolute_import
from .kademlia.node import Node as KademliaNode
from .node import EntangledNode
from .dtuple import DistributedTupleSpacePeer



#!/usr/bin/env python
# identity_restorer.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (id_restorer.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: id_restorer.

.. role:: red


EVENTS:
    * :red:`my-id-failed`
    * :red:`my-id-received`
    * :red:`restore-failed`
    * :red:`restore-success`
    * :red:`start`
    * :red:`stun-failed`
    * :red:`stun-success`
    * :red:`suppliers-read-failed`
    * :red:`suppliers-read-ok`

.. raw:: html

    <a href="https://bitdust.io/automats/identity_restorer/identity_restorer.png" target="_blank">
    <img src="https://bitdust.io/automats/identity_restorer/identity_restorer.png" style="max-width:100%;">
    </a>

A state machine to restore the user account.

Needed for restoration of the user account information using its Private key and ID url.
    * at first request user's identity file from Identity server
    * do verification and restoration of his locally identity to be able to start the software
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os
import sys
import random

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in identity_restorer.py')

#------------------------------------------------------------------------------

from bitdust.automats import automat

from bitdust.logs import lg

from bitdust.main import settings

from bitdust.system import bpio

from bitdust.lib import net_misc

from bitdust.raid import eccmap

from bitdust.crypt import key

from bitdust.contacts import contactsdb

from bitdust.stun import stun_client

from bitdust.dht import dht_relations

from bitdust.userid import identity
from bitdust.userid import my_id
from bitdust.userid import id_url

#------------------------------------------------------------------------------

_IdRestorer = None
_WorkingIDURL = ''
_WorkingKey = ''

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _IdRestorer
    if _IdRestorer is None:
        # set automat name and starting state here
        _IdRestorer = IdRestorer(
            name='id_restorer',
            state='AT_STARTUP',
            debug_level=2,
            log_transitions=True,
            log_events=True,
            publish_events=True,
        )
    if event is not None:
        _IdRestorer.automat(event, *args, **kwargs)
    return _IdRestorer


class IdRestorer(automat.Automat):

    """
    BitDust identity_restorer() Automat.

    Class to run the process to restore user account.
    """

    MESSAGES = {
        'MSG_01': ['requesting user identity from remote ID server'],
        'MSG_02': ['key verification failed!', 'red'],
        'MSG_03': ['downloading user identity from remote ID server'],
        'MSG_04': ['incorrect or non-existing IDURL provided', 'red'],
        'MSG_05': ['verifying user identity and private key'],
        'MSG_06': ['your identity restored successfully!', 'green'],
        'MSG_07': ['checking network connectivity'],
        'MSG_08': [
            'network connection failed',
            'red',
        ],
        'MSG_09': ['reading list of my suppliers from DHT'],
    }

    def init(self):
        self.last_message = ''

    def msg(self, msgid, *args, **kwargs):
        msg = self.MESSAGES.get(msgid, ['', 'black'])
        text = msg[0]
        color = 'black'
        if len(msg) == 2:
            color = msg[1]
        return text, color

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        from bitdust.main import installer
        installer.A('id_restorer.state', newstate)

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'start':
                self.state = 'STUN_MY_IP'
                self.doPrint(self.msg('MSG_07', *args, **kwargs))
                self.doSetWorkingIDURL(*args, **kwargs)
                self.doSetWorkingKey(*args, **kwargs)
                self.doStunExternalIP(*args, **kwargs)
        #---STUN_MY_IP---
        elif self.state == 'STUN_MY_IP':
            if event == 'stun-failed':
                self.state = 'FAILED'
                self.doPrint(self.msg('MSG_08', *args, **kwargs))
                self.doClearWorkingIDURL(*args, **kwargs)
                self.doClearWorkingKey(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stun-success':
                self.state = 'MY_ID'
                self.doPrint(self.msg('MSG_01', *args, **kwargs))
                self.doRequestMyIdentity(*args, **kwargs)
        #---MY_ID---
        elif self.state == 'MY_ID':
            if event == 'my-id-received':
                self.state = 'VERIFY'
                self.doPrint(self.msg('MSG_05', *args, **kwargs))
                self.doVerifyAndRestore(*args, **kwargs)
            elif event == 'my-id-failed':
                self.state = 'FAILED'
                self.doPrint(self.msg('MSG_04', *args, **kwargs))
                self.doClearWorkingIDURL(*args, **kwargs)
                self.doClearWorkingKey(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---VERIFY---
        elif self.state == 'VERIFY':
            if event == 'restore-failed':
                self.state = 'FAILED'
                self.doPrint(*args, **kwargs)
                self.doClearWorkingIDURL(*args, **kwargs)
                self.doClearWorkingKey(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'restore-success':
                self.state = 'SUPPLIERS?'
                self.doPrint(self.msg('MSG_09', *args, **kwargs))
                self.doDHTReadMySuppliers(*args, **kwargs)
        #---SUPPLIERS?---
        elif self.state == 'SUPPLIERS?':
            if event == 'suppliers-read-ok' or event == 'suppliers-read-failed':
                self.state = 'RESTORED!'
                self.doPrint(self.msg('MSG_06', *args, **kwargs))
                self.doRestoreSave(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---RESTORED!---
        elif self.state == 'RESTORED!':
            pass
        #---FAILED---
        elif self.state == 'FAILED':
            pass
        return None

    def doStunExternalIP(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'identity_restorer.doStunExternalIP')

        def save(result):
            if _Debug:
                lg.out(_DebugLevel, '            external IP : %s' % result)
            if result['result'] != 'stun-success':
                self.automat('stun-failed')
                return
            ip = result['ip']
            bpio.WriteTextFile(settings.ExternalIPFilename(), ip)
            self.automat('stun-success', ip)

        rnd_udp_port = random.randint(
            settings.DefaultUDPPort(),
            settings.DefaultUDPPort() + 500,
        )
        rnd_dht_port = random.randint(
            settings.DefaultDHTPort(),
            settings.DefaultDHTPort() + 500,
        )
        d = stun_client.safe_stun(udp_port=rnd_udp_port, dht_port=rnd_dht_port)
        d.addCallback(save)
        d.addErrback(lambda _: self.automat('stun-failed'))

    def doSetWorkingIDURL(self, *args, **kwargs):
        global _WorkingIDURL
        _WorkingIDURL = args[0]['idurl']

    def doSetWorkingKey(self, *args, **kwargs):
        global _WorkingKey
        _WorkingKey = args[0]['keysrc']

    def doClearWorkingIDURL(self, *args, **kwargs):
        global _WorkingIDURL
        _WorkingIDURL = ''

    def doClearWorkingKey(self, *args, **kwargs):
        global _WorkingKey
        _WorkingKey = ''

    def doRequestMyIdentity(self, *args, **kwargs):
        global _WorkingIDURL
        idurl = _WorkingIDURL
        if _Debug:
            lg.out(_DebugLevel, 'identity_restorer.doRequestMyIdentity %s %s' % (idurl, type(idurl)))
        net_misc.getPageTwisted(idurl).addCallbacks(lambda src: self.automat('my-id-received', src), lambda err: self.automat('my-id-failed', err))

    def doVerifyAndRestore(self, *args, **kwargs):
        global _WorkingKey
        if _Debug:
            lg.out(_DebugLevel, 'identity_restorer.doVerifyAndRestore')

        remote_identity_src = args[0]

        if os.path.isfile(settings.KeyFileName()):
            if _Debug:
                lg.out(_DebugLevel, 'identity_restorer.doVerifyAndRestore will backup and remove ' + settings.KeyFileName())
            bpio.backup_and_remove(settings.KeyFileName())

        if os.path.isfile(settings.LocalIdentityFilename()):
            if _Debug:
                lg.out(_DebugLevel, 'identity_restorer.doVerifyAndRestore will backup and remove ' + settings.LocalIdentityFilename())
            bpio.backup_and_remove(settings.LocalIdentityFilename())

        try:
            remote_ident = identity.identity(xmlsrc=remote_identity_src)
            local_ident = identity.identity(xmlsrc=remote_identity_src)
        except:
            # lg.exc()
            reactor.callLater(0.1, self.automat, 'restore-failed', ('remote identity have incorrect format', 'red'))  # @UndefinedVariable
            return

        if _Debug:
            lg.out(_DebugLevel, 'identity_restorer.doVerifyAndRestore checking remote identity')
        try:
            res = remote_ident.isCorrect()
        except:
            lg.exc()
            res = False
        if not res:
            if _Debug:
                lg.out(_DebugLevel, 'identity_restorer.doVerifyAndRestore remote identity is not correct FAILED!!!!')
            reactor.callLater(0.1, self.automat, 'restore-failed', ('remote identity format is not correct', 'red'))  # @UndefinedVariable
            return

        if _Debug:
            lg.out(_DebugLevel, 'identity_restorer.doVerifyAndRestore validate remote identity')
        try:
            res = remote_ident.Valid()
        except:
            lg.exc()
            res = False
        if not res:
            if _Debug:
                lg.out(_DebugLevel, 'identity_restorer.doVerifyAndRestore validate remote identity FAILED!!!!')
            reactor.callLater(0.1, self.automat, 'restore-failed', ('remote identity is not valid', 'red'))  # @UndefinedVariable
            return

        my_id.forgetLocalIdentity()
        my_id.eraseLocalIdentity(do_backup=True)
        key.ForgetMyKey(erase_file=True, do_backup=True)
        bpio.WriteTextFile(settings.KeyFileName(), _WorkingKey)
        try:
            key.InitMyKey()
        except:
            key.ForgetMyKey(erase_file=True, do_backup=False)
            reactor.callLater(0.1, self.automat, 'restore-failed', ('private key is not valid', 'red'))  # @UndefinedVariable
            return

        try:
            local_ident.sign()
        except:
            # lg.exc()
            reactor.callLater(0.1, self.automat, 'restore-failed', ('error while signing identity', 'red'))  # @UndefinedVariable
            return

        if remote_ident.signature != local_ident.signature:
            reactor.callLater(0.1, self.automat, 'restore-failed', ('signature did not match, key verification failed!', 'red'))  # @UndefinedVariable
            return

        my_id.setLocalIdentity(local_ident)
        my_id.saveLocalIdentity()
        # my_id.rebuildLocalIdentity(save_identity=True)

        bpio.WriteTextFile(settings.UserNameFilename(), my_id.getIDName())

        if os.path.isfile(settings.KeyFileName() + '.backup'):
            if _Debug:
                lg.out(_DebugLevel, 'identity_restorer.doVerifyAndRestore will remove backup file for ' + settings.KeyFileName())
            bpio.remove_backuped_file(settings.KeyFileName())

        if os.path.isfile(settings.LocalIdentityFilename() + '.backup'):
            if _Debug:
                lg.out(_DebugLevel, 'identity_restorer.doVerifyAndRestore will remove backup file for ' + settings.LocalIdentityFilename())
            bpio.remove_backuped_file(settings.LocalIdentityFilename())

        reactor.callLater(0.1, self.automat, 'restore-success')  # @UndefinedVariable

    def doDHTReadMySuppliers(self, *args, **kwargs):
        """
        Action method.
        """
        known_suppliers = len(contactsdb.suppliers())
        if known_suppliers > 0:
            lg.warn('skip reading my suppliers from DHT, currently known %d suppliers already' % known_suppliers)
            self.automat('suppliers-read-ok')
            return
        d = dht_relations.read_customer_suppliers(my_id.getIDURL(), use_cache=False)
        d.addCallback(self._on_my_dht_relations_discovered)
        d.addErrback(self._on_my_dht_relations_failed)

    def doRestoreSave(self, *args, **kwargs):
        """
        TODO: use lib.config here request settings from DHT my suppliers need
        to keep that settings in DHT.
        """
        # settings.uconfig().set('storage.suppliers', '0')
        # settings.uconfig().set('storage.needed', '0Mb')
        # settings.uconfig().set('storage.donated', '0Mb')
        # settings.uconfig().update()

    def doPrint(self, *args, **kwargs):
        from bitdust.main import installer
        installer.A().event('print', args[0])
        self.last_message = args[0][0]
        if _Debug:
            lg.out(_DebugLevel, 'id_restorer.doPrint: %s' % str(args[0]))

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        self.executeStateChangedCallbacks(oldstate=None, newstate=self.state, event_string=None, args=args)
        self.destroy(dead_state=self.state)
        global _IdRestorer
        _IdRestorer = None

    def _on_my_dht_relations_discovered(self, dht_result):
        if not (dht_result and isinstance(dht_result, dict) and len(dht_result.get('suppliers', [])) > 0):
            lg.warn('no dht records found for my customer family')
            self.automat('suppliers-read-failed')
            return
        dht_suppliers = id_url.to_bin_list(dht_result['suppliers'])
        dht_ecc_map = dht_result.get('ecc_map', settings.DefaultEccMapName())
        try:
            dht_desired_suppliers_number = eccmap.GetEccMapSuppliersNumber(dht_ecc_map)
        except:
            lg.exc()
            dht_desired_suppliers_number = eccmap.GetEccMapSuppliersNumber(settings.DefaultEccMapName())
        settings.config.conf().setInt('services/customer/suppliers-number', dht_desired_suppliers_number)
        contactsdb.set_suppliers(dht_suppliers)
        contactsdb.save_suppliers()
        lg.info('found and restored list of %d suppliers from DHT' % dht_desired_suppliers_number)
        self.automat('suppliers-read-ok')

    def _on_my_dht_relations_failed(self, err):
        lg.err(err)
        self.automat('suppliers-read-failed')



#!/usr/bin/python
# service_list_files.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_list_files.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_list_files
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return ListFilesService()


class ListFilesService(LocalService):

    service_name = 'service_list_files'
    config_path = 'services/list-files/enabled'

    def dependent_on(self):
        return [
            'service_employer',
        ]

    def init(self, **kwargs):
        from bitdust.main import events
        events.add_subscriber(self._on_my_suppliers_all_hired, 'my-suppliers-all-hired')
        events.add_subscriber(self._on_my_suppliers_yet_not_hired, 'my-suppliers-yet-not-hired')

    def shutdown(self):
        from bitdust.main import events
        events.remove_subscriber(self._on_my_suppliers_yet_not_hired, 'my-suppliers-yet-not-hired')
        events.remove_subscriber(self._on_my_suppliers_all_hired, 'my-suppliers-all-hired')

    def start(self):
        from bitdust.logs import lg
        from bitdust.customer import fire_hire
        if not fire_hire.IsAllHired():
            lg.warn('service_list_files() can not start right now, not all suppliers hired yet')
            return False
        from bitdust.customer import list_files_orator
        list_files_orator.A('init')
        return True

    def stop(self):
        from bitdust.customer import list_files_orator
        list_files_orator.Destroy()
        return True

    def _on_my_suppliers_all_hired(self, evt):
        from bitdust.logs import lg
        from bitdust.services import driver
        if driver.is_enabled('service_list_files'):
            if not driver.is_started('service_list_files'):
                lg.info('all my suppliers are hired, starting service_list_files()')
                driver.start_single('service_list_files')
            from bitdust.customer import list_files_orator
            if list_files_orator.A():
                list_files_orator.synchronize_files()

    def _on_my_suppliers_yet_not_hired(self, evt):
        from bitdust.logs import lg
        from bitdust.services import driver
        if driver.is_enabled('service_list_files'):
            if driver.is_started('service_list_files'):
                lg.info('my suppliers failed to hire, stopping service_list_files()')
                driver.stop_single('service_list_files')



#!/usr/bin/python
# api_web_socket.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (api_web_socket.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: api_web_socket
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

_APILogFileEnabled = False

#------------------------------------------------------------------------------

from twisted.application.strports import listen
from twisted.internet.defer import Deferred
from twisted.internet.protocol import Protocol, Factory
from twisted.python.failure import Failure

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import txws
from bitdust.lib import serialization

from bitdust.system import local_fs

from bitdust.main import events
from bitdust.main import settings

from bitdust.interface import api

#------------------------------------------------------------------------------

_WebSocketListener = None
_WebSocketTransports = {}
_AllAPIMethods = []
_APISecret = None

#------------------------------------------------------------------------------


def init(port=None):
    global _WebSocketListener
    global _AllAPIMethods
    global _APILogFileEnabled
    _APILogFileEnabled = settings.config.conf().getBool('logs/api-enabled')
    if _WebSocketListener is not None:
        lg.warn('_WebSocketListener already initialized')
        return
    if not port:
        port = settings.DefaultWebSocketPort()
    try:
        ws = BitDistWrappedWebSocketFactory(BitDustWebSocketFactory())
        _WebSocketListener = listen('tcp:%d' % port, ws)
    except:
        lg.exc()
        return
    _AllAPIMethods = set(dir(api))
    _AllAPIMethods.difference_update(
        [
            # TODO: keep that list up to date when changing the api
            'on_api_result_prepared',
            'Deferred',
            'ERROR',
            'Failure',
            'OK',
            'RESULT',
            '_Debug',
            '_DebugLevel',
            'strng',
            'sys',
            'time',
            'gc',
            'map',
            'os',
            '__builtins__',
            '__cached__',
            '__doc__',
            '__file__',
            '__loader__',
            '__name__',
            '__package__',
            '__spec__',
            'absolute_import',
            'driver',
            'filemanager',
            'jsn',
            'lg',
            'event_listen',
            'message_receive',
        ]
    )
    if _Debug:
        lg.out(_DebugLevel, 'api_web_socket.init  _WebSocketListener=%r with %d methods' % (_WebSocketListener, len(_AllAPIMethods)))
    read_api_secret()
    events.add_subscriber(on_event, event_id='*')


def shutdown():
    global _WebSocketListener
    events.remove_subscriber(on_event, event_id='*')
    if _WebSocketListener:
        if _Debug:
            lg.out(_DebugLevel, 'api_web_socket.shutdown calling _WebSocketListener.stopListening()')
        _WebSocketListener.stopListening()
        del _WebSocketListener
        _WebSocketListener = None
        if _Debug:
            lg.out(_DebugLevel, '    _WebSocketListener destroyed')
    else:
        lg.warn('_WebSocketListener is None')


#------------------------------------------------------------------------------


def read_api_secret():
    global _APISecret
    _APISecret = local_fs.ReadTextFile(settings.APISecretFile())


#------------------------------------------------------------------------------


class BitDustWrappedWebSocketProtocol(txws.WebSocketProtocol):

    def validateHeaders(self):
        global _APISecret
        if _APISecret:
            access_granted = False
            _, _, api_secret_parameter = self.location.partition('?')
            param_name, _, api_secret_parameter = api_secret_parameter.partition('=')
            if param_name == 'a' or param_name == 'api_secret':
                if api_secret_parameter == _APISecret:
                    access_granted = True
            if not access_granted:
                events.send('web-socket-access-denied', data=dict())
                self.loseConnection()
                return
        return txws.WebSocketProtocol.validateHeaders(self)


class BitDistWrappedWebSocketFactory(txws.WebSocketFactory):

    protocol = BitDustWrappedWebSocketProtocol


#------------------------------------------------------------------------------


class BitDustWebSocketProtocol(Protocol):

    _key = None

    def dataReceived(self, data):
        try:
            json_data = serialization.BytesToDict(data, keys_to_text=True, values_to_text=True, encoding='utf-8')
        except:
            lg.exc()
            return
        if _Debug:
            lg.dbg(_DebugLevel, 'received %d bytes from web socket: %r' % (len(data), json_data))
        if not do_process_incoming_message(json_data):
            lg.warn('failed processing incoming message from web socket: %r' % json_data)

    def connectionMade(self):
        global _WebSocketTransports
        Protocol.connectionMade(self)
        peer = self.transport.getPeer()
        self._key = (
            peer.type,
            peer.host,
            peer.port,
        )
        peer = '%s://%s:%s' % (self._key[0], self._key[1], self._key[2])
        _WebSocketTransports[self._key] = self.transport
        if _Debug:
            lg.args(_DebugLevel, key=self._key, ws_connections=len(_WebSocketTransports))
        events.send('web-socket-connected', data=dict(peer=peer))

    def connectionLost(self, *args, **kwargs):
        global _WebSocketTransports
        if _Debug:
            lg.args(_DebugLevel, key=self._key, ws_connections=len(_WebSocketTransports))
        Protocol.connectionLost(self, *args, **kwargs)
        _WebSocketTransports.pop(self._key)
        peer = '%s://%s:%s' % (self._key[0], self._key[1], self._key[2])
        self._key = None
        events.send('web-socket-disconnected', data=dict(peer=peer))


#------------------------------------------------------------------------------


class BitDustWebSocketFactory(Factory):

    protocol = BitDustWebSocketProtocol

    def buildProtocol(self, addr):
        """
        Only accepting connections from local machine!
        """
        if addr.host != '127.0.0.1':
            lg.err('refused connection from remote host: %r' % addr.host)
            return None
        proto = Factory.buildProtocol(self, addr)
        return proto


#------------------------------------------------------------------------------


def do_process_incoming_message(json_data):
    global _AllAPIMethods
    command = json_data.get('command')
    if command == 'api_call':
        method = json_data.get('method', None)
        kwargs = json_data.get('kwargs', {})
        call_id = json_data.get('call_id', None)

        if not method:
            lg.warn('api method name was not provided')
            return push({
                'type': 'api_call',
                'payload': {
                    'call_id': call_id,
                    'errors': ['api method name was not provided'],
                },
            })

        if method not in _AllAPIMethods:
            lg.warn('invalid api method name: %r' % method)
            return push({
                'type': 'api_call',
                'payload': {
                    'call_id': call_id,
                    'errors': ['invalid api method name'],
                },
            })

        if _Debug:
            lg.out(_DebugLevel, '*** %s  API WS IN  %s(%r)' % (call_id, method, kwargs))

        if _APILogFileEnabled:
            lg.out(0, '*** %s  WS IN  %s(%r)' % (call_id, method, kwargs), log_name='api', showtime=True)

        func = getattr(api, method)
        try:
            response = func(**kwargs)
        except Exception as err:
            lg.err(f'{method}({kwargs}) : {err}')
            return push({
                'type': 'api_call',
                'payload': {
                    'call_id': call_id,
                    'errors': [str(err)],
                },
            })

        if isinstance(response, Deferred):

            def _cb(r):
                return push({
                    'type': 'api_call',
                    'payload': {
                        'call_id': call_id,
                        'response': r,
                    },
                })

            def _eb(err):
                err_msg = err.getErrorMessage() if isinstance(err, Failure) else str(err)
                return push({
                    'type': 'api_call',
                    'payload': {
                        'call_id': call_id,
                        'errors': [err_msg],
                    },
                })

            response.addCallback(_cb)
            response.addErrback(_eb)
            return True

        return push({
            'type': 'api_call',
            'payload': {
                'call_id': call_id,
                'response': response,
            },
        })

    return False


#------------------------------------------------------------------------------


def on_event(evt):
    return push({
        'type': 'event',
        'payload': {
            'event_id': evt.event_id,
            'data': evt.data,
        },
    })


def on_stream_message(message_json):
    return push({
        'type': 'stream_message',
        'payload': message_json,
    })


def on_online_status_changed(status_info):
    return push({
        'type': 'online_status',
        'payload': status_info,
    })


def on_model_changed(snapshot_object):
    return push({
        'type': 'model',
        'payload': snapshot_object.to_json(),
    })


#------------------------------------------------------------------------------


def push(json_data):
    global _WebSocketTransports
    if not _WebSocketTransports:
        return False
    raw_bytes = serialization.DictToBytes(json_data, encoding='utf-8')
    for _key, transp in _WebSocketTransports.items():
        try:
            transp.write(raw_bytes)
        except:
            lg.exc()
            continue
        if _Debug:
            lg.dbg(_DebugLevel, 'sent %d bytes to web socket %s' % (len(raw_bytes), '%s://%s:%s' % (_key[0], _key[1], _key[2])))
    if _Debug:
        lg.out(_DebugLevel, '***   API WS PUSH  %d bytes' % len(raw_bytes))
    if _APILogFileEnabled:
        lg.out(
            0,
            '*** WS PUSH  %d bytes : %r' % (
                len(raw_bytes),
                json_data,
            ),
            log_name='api',
            showtime=True,
        )
    return True
